import numpy

import reikna.helpers as helpers
import reikna.cluda.dtypes as dtypes
from reikna.cluda import Module

TEMPLATE = helpers.template_for(__file__)


def create_struct_types(word_dtype, key_words, counter_words):

    key_dtype = numpy.dtype([('v', (word_dtype, (key_words,)))])
    key_ctype = dtypes.ctype_module(key_dtype)

    counter_dtype = numpy.dtype([('v', (word_dtype, (counter_words,)))])
    counter_ctype = dtypes.ctype_module(counter_dtype)

    return key_dtype, key_ctype, counter_dtype, counter_ctype


class Bijection:
    """
    Contains a CBRNG bijection module and accompanying metadata.
    Supports ``__process_modules__`` protocol.

    .. py:attribute:: word_dtype

        The data type of the integer word used by the generator.

    .. py:attribute:: key_words

        The number of words used by the key.

    .. py:attribute:: counter_words

        The number of words used by the counter.

    .. py:attribute:: key_dtype

        The ``numpy.dtype`` object representing a bijection key.
        Contains a single array field ``v`` with ``key_words`` of ``word_dtype`` elements.

    .. py:attribute:: counter_dtype

        The ``numpy.dtype`` object representing a bijection counter.
        Contains a single array field ``v`` with ``key_words`` of ``word_dtype`` elements.

    .. py:attribute:: raw_functions

        A dictionary ``dtype:function_name`` of available functions ``function_name``
        in :py:attr:`module` that produce a random full-range integer ``dtype``
        from a :c:type:`State`, advancing it.
        Available functions: :c:func:`get_raw_uint32`, :c:func:`get_raw_uint64`.

    .. py:attribute:: module

        The module containing the CBRNG function.
        It provides the C functions below.

    .. c:macro:: COUNTER_WORDS

        Contains the value of :py:attr:`counter_words`.

    .. c:macro:: KEY_WORDS

        Contains the value of :py:attr:`key_words`.

    .. c:type:: Word

        Contains the type corresponding to :py:attr:`word_dtype`.

    .. c:type:: Key

        Describes the bijection key.
        Alias for the structure generated from :py:attr:`key_dtype`.

        .. c:member:: Word v[KEY_WORDS]

    .. c:type:: Counter

        Describes the bijection counter, or its output.
        Alias for the structure generated from :py:attr:`counter_dtype`.

        .. c:member:: Word v[COUNTER_WORDS]

    .. c:function:: Counter make_counter_from_int(int x)

        Creates a counter object from an integer.

    .. c:function:: Counter bijection(Key key, Counter counter)

        The main bijection function.

    .. c:type:: State

        A structure containing the CBRNG state which is used by :py:mod:`~reikna.cbrng.samplers`.

    .. c:function:: State make_state(Key key, Counter counter)

        Creates a new state object.

    .. c:function:: Counter get_next_unused_counter(State state)

        Extracts a counter which has not been used in random sampling.

    .. c:type:: uint32

        A type of unsigned 32-bit word, corresponds to ``numpy.uint32``.

    .. c:type:: uint64

        A type of unsigned 64-bit word, corresponds to ``numpy.uint64``.

    .. c:function:: uint32 get_raw_uint32(State *state)

        Returns uniformly distributed unsigned 32-bit word and updates the state.

    .. c:function:: uint64 get_raw_uint64(State *state)

        Returns uniformly distributed unsigned 64-bit word and updates the state.
    """
    def __init__(self, module, word_dtype, key_dtype, counter_dtype):
        """__init__()""" # hide the signature from Sphinx

        self.module = module
        self.word_dtype = word_dtype

        self.key_words = key_dtype.fields['v'][0].shape[0]
        self.counter_words = counter_dtype.fields['v'][0].shape[0]

        self.counter_dtype = counter_dtype
        self.key_dtype = key_dtype

        # Compensate for the mysterious distinction numpy makes between
        # a predefined dtype and a generic dtype.
        self.raw_functions = {
            numpy.uint32: 'get_raw_uint32',
            numpy.dtype('uint32'): 'get_raw_uint32',
            numpy.uint64: 'get_raw_uint64',
            numpy.dtype('uint64'): 'get_raw_uint64',
        }

    def __process_modules__(self, process):
        return Bijection(
            process(self.module), self.word_dtype, self.key_dtype, self.counter_dtype)


def threefry(bitness, counter_words, rounds=20):
    """
    A CBRNG based on a big number of fast rounds (bit rotations).

    :param bitness: ``32`` or ``64``, corresponds to the size of generated random integers.
    :param counter_words: ``2`` or ``4``, number of integers generated in one go.
    :param rounds: ``1`` to ``72``, the more rounds, the better randomness is achieved.
        Default values are big enough to qualify as PRNG.
    :returns: a :py:class:`Bijection` object.
    """

    ROTATION_CONSTANTS = {
        # These are the R_256 constants from the Threefish reference sources
        # with names changed to R_64x4...
        (64, 4): numpy.array([[14, 52, 23, 5, 25, 46, 58, 32], [16, 57, 40, 37, 33, 12, 22, 32]]).T,

        # Output from skein_rot_search: (srs64_B64-X1000)
        # Random seed = 1. BlockSize = 128 bits. sampleCnt =  1024. rounds =  8, minHW_or=57
        # Start: Tue Mar  1 10:07:48 2011
        # rMin = 0.136. #0325[*15] [CRC=455A682F. hw_OR=64. cnt=16384. blkSize= 128].format
        (64, 2): numpy.array([[16, 42, 12, 31, 16, 32, 24, 21]]).T,
        # 4 rounds: minHW =  4  [  4  4  4  4 ]
        # 5 rounds: minHW =  8  [  8  8  8  8 ]
        # 6 rounds: minHW = 16  [ 16 16 16 16 ]
        # 7 rounds: minHW = 32  [ 32 32 32 32 ]
        # 8 rounds: minHW = 64  [ 64 64 64 64 ]
        # 9 rounds: minHW = 64  [ 64 64 64 64 ]
        # 10 rounds: minHW = 64  [ 64 64 64 64 ]
        # 11 rounds: minHW = 64  [ 64 64 64 64 ]

        # Output from skein_rot_search: (srs-B128-X5000.out)
        # Random seed = 1. BlockSize = 64 bits. sampleCnt =  1024. rounds =  8, minHW_or=28
        # Start: Mon Aug 24 22:41:36 2009
        # ...
        # rMin = 0.472. #0A4B[*33] [CRC=DD1ECE0F. hw_OR=31. cnt=16384. blkSize= 128].format
        (32, 4): numpy.array([[10, 11, 13, 23, 6, 17, 25, 18], [26, 21, 27, 5, 20, 11, 10, 20]]).T,
        # 4 rounds: minHW =  3  [  3  3  3  3 ]
        # 5 rounds: minHW =  7  [  7  7  7  7 ]
        # 6 rounds: minHW = 12  [ 13 12 13 12 ]
        # 7 rounds: minHW = 22  [ 22 23 22 23 ]
        # 8 rounds: minHW = 31  [ 31 31 31 31 ]
        # 9 rounds: minHW = 32  [ 32 32 32 32 ]
        # 10 rounds: minHW = 32  [ 32 32 32 32 ]
        # 11 rounds: minHW = 32  [ 32 32 32 32 ]

        # Output from skein_rot_search (srs32x2-X5000.out)
        # Random seed = 1. BlockSize = 64 bits. sampleCnt =  1024. rounds =  8, minHW_or=28
        # Start: Tue Jul 12 11:11:33 2011
        # rMin = 0.334. #0206[*07] [CRC=1D9765C0. hw_OR=32. cnt=16384. blkSize=  64].format
        (32, 2): numpy.array([[13, 15, 26, 6, 17, 29, 16, 24]]).T
        # 4 rounds: minHW =  4  [  4  4  4  4 ]
        # 5 rounds: minHW =  6  [  6  8  6  8 ]
        # 6 rounds: minHW =  9  [  9 12  9 12 ]
        # 7 rounds: minHW = 16  [ 16 24 16 24 ]
        # 8 rounds: minHW = 32  [ 32 32 32 32 ]
        # 9 rounds: minHW = 32  [ 32 32 32 32 ]
        # 10 rounds: minHW = 32  [ 32 32 32 32 ]
        # 11 rounds: minHW = 32  [ 32 32 32 32 ]
    }

    # Taken from Skein
    PARITY_CONSTANTS = {
        64: numpy.uint64(0x1BD11BDAA9FC1A22),
        32: numpy.uint32(0x1BD11BDA)
    }

    assert 1 <= rounds <= 72

    word_dtype = dtypes.normalize_type(numpy.uint32 if bitness == 32 else numpy.uint64)
    key_words = counter_words
    key_dtype, key_ctype, counter_dtype, counter_ctype = create_struct_types(
        word_dtype, key_words, counter_words)

    module = Module(
        TEMPLATE.get_def("threefry"),
        render_kwds=dict(
            word_dtype=word_dtype, word_ctype=dtypes.ctype(word_dtype),
            key_words=key_words, counter_words=counter_words,
            key_ctype=key_ctype, counter_ctype=counter_ctype,
            rounds=rounds, rotation_constants=ROTATION_CONSTANTS[(bitness, counter_words)],
            parity_constant=PARITY_CONSTANTS[bitness]))

    return Bijection(module, word_dtype, key_dtype, counter_dtype)


def philox(bitness, counter_words, rounds=10):
    """
    A CBRNG based on a low number of slow rounds (multiplications).

    :param bitness: ``32`` or ``64``, corresponds to the size of generated random integers.
    :param counter_words: ``2`` or ``4``, number of integers generated in one go.
    :param rounds: ``1`` to ``12``, the more rounds, the better randomness is achieved.
        Default values are big enough to qualify as PRNG.
    :returns: a :py:class:`Bijection` object.
    """

    W_CONSTANTS = {
        64: [
            numpy.uint64(0x9E3779B97F4A7C15), # golden ratio
            numpy.uint64(0xBB67AE8584CAA73B) # sqrt(3)-1
        ],
        32: [
            numpy.uint32(0x9E3779B9), # golden ratio
            numpy.uint32(0xBB67AE85) # sqrt(3)-1
        ]
    }

    M_CONSTANTS = {
        (64,2): [numpy.uint64(0xD2B74407B1CE6E93)],
        (64,4): [numpy.uint64(0xD2E7470EE14C6C93), numpy.uint64(0xCA5A826395121157)],
        (32,2): [numpy.uint32(0xD256D193)],
        (32,4): [numpy.uint32(0xD2511F53), numpy.uint32(0xCD9E8D57)]
    }

    assert 1 <= rounds <= 12
    word_dtype = dtypes.normalize_type(numpy.uint32 if bitness == 32 else numpy.uint64)
    key_words = counter_words // 2
    key_dtype, key_ctype, counter_dtype, counter_ctype = create_struct_types(
        word_dtype, key_words, counter_words)

    module = Module(
        TEMPLATE.get_def("philox"),
        render_kwds=dict(
            word_dtype=word_dtype, word_ctype=dtypes.ctype(word_dtype),
            key_words=key_words, counter_words=counter_words,
            key_ctype=key_ctype, counter_ctype=counter_ctype,
            rounds=rounds, w_constants=W_CONSTANTS[bitness],
            m_constants=M_CONSTANTS[(bitness, counter_words)]))

    return Bijection(module, word_dtype, key_dtype, counter_dtype)



import time
import numpy
import pytest

from helpers import *

from reikna.matrixmul import MatrixMul
import reikna.cluda.dtypes as dtypes
from reikna.cluda import OutOfResourcesError
from reikna.helpers import product


def pytest_generate_tests(metafunc):

    if 'perf_bwo' in metafunc.funcargnames:
        bwos = [8, 16, 32]
        ids=["8x8", "16x16", "32x32"]
        metafunc.parametrize('perf_bwo', bwos, ids=ids)

    if 'perf_shapes' in metafunc.funcargnames:

        shapes = []
        for s in [(64, 64), (256, 256), (1024, 1024)]:
            shapes.append((s, s))

        for s in [(1024, 16, 16), (256, 32, 32), (512, 25, 25)]:
            shapes.append((s, s[1:]))
            shapes.append((s, s))

        ids = []
        for s in shapes:
            ids.append(str(s[0]) + 'x' + str(s[1]))

        metafunc.parametrize('perf_shapes', shapes, ids=ids)

    if 'shapes' in metafunc.funcargnames:

        shapes = [
            ((15, 17), (17, 3)),
            ((122, 5), (5, 1000)),
            ((45, 99), (30, 99, 40)),
            ((12, 56, 78), (12, 78, 44))
        ]

        ids = [str(s1) + 'x' + str(s2) for s1, s2 in shapes]

        metafunc.parametrize('shapes', shapes, ids=ids)

    if 'arg_dtypes' in metafunc.funcargnames:

        arg_dtypes = [(False, False), (False, True), (True, False), (True, True)]
        mark = lambda x: 'c' if x else 'r'
        ids = [mark(t1) + mark(t2) for t1, t2 in arg_dtypes]
        metafunc.parametrize('arg_dtypes', arg_dtypes, ids=ids)


def ref_dot(a, b):
    a_batch = product(a.shape[:-2])
    b_batch = product(b.shape[:-2])

    assert a_batch == b_batch or a_batch == 1 or b_batch == 1

    a = a.reshape(a_batch, a.shape[-2], a.shape[-1])
    b = b.reshape(b_batch, b.shape[-2], b.shape[-1])

    out_batch = max(a_batch, b_batch)
    out_shape = (out_batch, a.shape[-2], b.shape[-1])
    out_dtype = numpy.result_type(a.dtype, b.dtype)
    out = numpy.empty(out_shape, out_dtype)

    for i in range(out_batch):
        ai = 0 if a_batch == 1 else i
        bi = 0 if b_batch == 1 else i
        out[i] = numpy.dot(a[ai], b[bi])

    if a_batch == b_batch == 1:
        out = out.reshape(out.shape[-2], out_shape[-1])

    return out


def test_errors(thr_and_double, shapes, arg_dtypes):

    thr, double = thr_and_double
    s1, s2 = shapes
    c1, c2 = arg_dtypes

    dtype = numpy.float64 if double else numpy.float32
    dtype1 = dtypes.complex_for(dtype) if c1 else dtype
    dtype2 = dtypes.complex_for(dtype) if c2 else dtype

    a = get_test_array(s1, dtype1)
    b = get_test_array(s2, dtype2)
    res_ref = ref_dot(a, b)

    a_dev = thr.to_device(a)
    b_dev = thr.to_device(b)
    res_dev = thr.empty_like(res_ref)

    dot = MatrixMul(a_dev, b_dev, out_arr=res_dev)
    dotc = dot.compile(thr)
    dotc(res_dev, a_dev, b_dev)

    assert diff_is_negligible(res_dev.get(), res_ref)


def check_performance(thr_and_double, shape1, shape2, bwo):

    thr, double = thr_and_double
    dtype = numpy.float64 if double else numpy.float32

    a = get_test_array(shape1, dtype)
    b = get_test_array(shape2, dtype)

    a_dev = thr.to_device(a)
    b_dev = thr.to_device(b)
    res_ref = ref_dot(a, b)
    res_dev = thr.array(res_ref.shape, dtype=dtype)

    dot = MatrixMul(a_dev, b_dev, out_arr=res_dev, block_width_override=bwo)

    try:
        dotc = dot.compile(thr)
    except ValueError:
        pytest.skip()

    attempts = 10
    t1 = time.time()
    for i in range(attempts):
        dotc(res_dev, a_dev, b_dev)
    thr.synchronize()
    t2 = time.time()

    assert diff_is_negligible(thr.from_device(res_dev), res_ref)

    return (t2 - t1) / attempts, product(res_ref.shape) * shape1[-1] * 2


@pytest.mark.perf
@pytest.mark.returns('GFLOPS')
def test_performance_shape(thr_and_double, perf_shapes):
    return check_performance(thr_and_double, perf_shapes[0], perf_shapes[1], None)


@pytest.mark.perf
@pytest.mark.returns('GFLOPS')
def test_performance_block_width(thr_and_double, perf_bwo):
    return check_performance(thr_and_double, (512, 512), (512, 512), perf_bwo)



import numpy
import pytest

import reikna.cluda as cluda
import reikna.cluda.dtypes as dtypes


def flatten_dtype(dtype, prefix=[]):

    if dtype.names is None:
        return [(prefix, dtype)]
    else:
        result = []
        for name in dtype.names:
            nested_dtype, offset = dtype.fields[name]
            result += flatten_dtype(nested_dtype, prefix=prefix + [name])
        return result


def extract_field(arr, path):
    if len(path) == 0:
        return arr
    else:
        return extract_field(arr[path[0]], path[1:])


def check_struct_fill(thr, dtype):
    """
    Fill every field of the given ``dtype`` with its number and check the results.
    This helps to detect issues with alignment in the struct.
    """
    struct = dtypes.ctype_module(dtype)

    program = thr.compile(
    """
    KERNEL void test(GLOBAL_MEM ${struct} *dest)
    {
      const SIZE_T i = get_global_id(0);
      ${struct} res;

      %for i, field_info in enumerate(flatten_dtype(dtype)):
      res.${".".join(field_info[0])} = ${i};
      %endfor

      dest[i] = res;
    }
    """, render_kwds=dict(
        struct=struct,
        dtype=dtype,
        flatten_dtype=flatten_dtype))

    test = program.test

    a_dev = thr.array(128, dtype)
    test(a_dev, global_size=128)
    a = a_dev.get()

    for i, field_info in enumerate(flatten_dtype(dtype)):
        path, _ = field_info
        assert (extract_field(a, path) == i).all()


def test_hardcoded_alignment(thr):
    """
    Test the correctness of alignment for an explicit set of field offsets.
    """

    dtype_nested = numpy.dtype(dict(
        names=['val1', 'pad'],
        formats=[numpy.int32, numpy.int8],
        offsets=[0, 4],
        itemsize=8))

    dtype = numpy.dtype(dict(
        names=['val1', 'val2', 'nested'],
        formats=[numpy.int32, numpy.int16, dtype_nested],
        offsets=[0, 4, 8],
        itemsize=32))

    check_struct_fill(thr, dtype)


def test_adjusted_alignment(thr):
    """
    Test the correctness of alignment for field offsets adjusted automatically.
    """

    dtype_nested = numpy.dtype([
        ('val1', numpy.int32), ('pad', numpy.int8)])

    dtype = numpy.dtype([
        ('val1', numpy.int32),
        ('val2', numpy.int16),
        ('nested', dtype_nested)])

    dtype = dtypes.adjust_alignment(thr, dtype)

    check_struct_fill(thr, dtype)


def test_nested_array(thr):
    """
    Check that structures with nested arrays are processed correctly.
    """
    dtype_nested = numpy.dtype(dict(
        names=['val1', 'pad'],
        formats=[numpy.int8, numpy.int8]))

    dtype = numpy.dtype(dict(
        names=['pad', 'struct_arr', 'regular_arr'],
        formats=[numpy.int32, numpy.dtype((dtype_nested, 2)), numpy.dtype((numpy.int16, 3))]))

    dtype = dtypes.adjust_alignment(thr, dtype)
    struct = dtypes.ctype_module(dtype)

    program = thr.compile(
    """
    KERNEL void test(GLOBAL_MEM ${struct} *dest)
    {
      const SIZE_T i = get_global_id(0);
      ${struct} res;

      res.struct_arr[0].val1 = i + 0;
      res.struct_arr[1].val1 = i + 1;

      res.regular_arr[0] = i + 2;
      res.regular_arr[1] = i + 3;
      res.regular_arr[2] = i + 4;

      dest[i] = res;
    }
    """, render_kwds=dict(
        struct=struct,
        dtype=dtype,
        flatten_dtype=flatten_dtype))

    test = program.test

    a_dev = thr.array(64, dtype)
    test(a_dev, global_size=64)
    a = a_dev.get()

    idxs = numpy.arange(64)
    assert (a['struct_arr'][:,0]['val1'] == idxs + 0).all()
    assert (a['struct_arr'][:,1]['val1'] == idxs + 1).all()
    assert (a['regular_arr'][:,0] == idxs + 2).all()
    assert (a['regular_arr'][:,1] == idxs + 3).all()
    assert (a['regular_arr'][:,2] == idxs + 4).all()


def test_structural_typing(some_thr):
    """
    Checks that ``ctype_module`` for equal dtype objects result in the same module object
    (which means that these two types will actually be rendered as a single type).
    """
    dtype = numpy.dtype([('val1', numpy.int32), ('val2', numpy.float32)])

    struct1 = dtypes.ctype_module(dtype)
    struct2 = dtypes.ctype_module(dtype)

    # Check that these reference the same object
    assert struct1 is struct2

    # Just in case, compile a test kernel
    program = some_thr.compile(
    """
    KERNEL void test(GLOBAL_MEM float *dest)
    {
      const SIZE_T i = get_global_id(0);
      ${struct1} temp1;
      ${struct2} temp2;

      temp1.val1 = 0;
      temp1.val2 = 1;

      // If struct1 and struct2 correspond to different types,
      // this will give a compilation error,
      // because C has a nominative typing system.
      temp2 = temp1;

      dest[i] = temp2.val1 + temp2.val2;
    }
    """, render_kwds=dict(
        struct1=struct1,
        struct2=struct2))


def test_structural_typing_nested(some_thr):
    """
    Check that the structural typing behavior works for nested structures.
    In other words, a nested dtype gets represented by the same module as
    an equal top-level dtype.
    """
    dtype_nested = numpy.dtype([('val1', numpy.int32), ('val2', numpy.float32)])
    dtype = numpy.dtype([
        ('val1', numpy.int32), ('val2', numpy.float32),
        ('nested', dtype_nested)])

    struct_nested = dtypes.ctype_module(dtype_nested)
    struct = dtypes.ctype_module(dtype)

    program = some_thr.compile(
    """
    KERNEL void test(GLOBAL_MEM float *dest)
    {
      const SIZE_T i = get_global_id(0);
      ${struct_nested} temp_nested;
      ${struct} temp;

      temp_nested.val1 = 0;
      temp_nested.val2 = 1;

      // If the nested structure has a different type from temp_nested,
      // this will give a compilation error.
      temp.nested = temp_nested;
      temp.val1 = 0;
      temp.val2 = 1;

      dest[i] = temp.val1 + temp.val2 + temp.nested.val1 + temp.nested.val2;
    }
    """, render_kwds=dict(
        struct=struct,
        struct_nested=struct_nested))



import numpy

import reikna.helpers as helpers
from reikna.core import Computation, Parameter, Annotation, Type
from reikna.cbrng.tools import KeyGenerator
from reikna.cbrng.bijections import philox
from reikna.cbrng.samplers import SAMPLERS

TEMPLATE = helpers.template_for(__file__)


class CBRNG(Computation):
    """
    Bases: :py:class:`~reikna.core.Computation`

    Counter-based pseudo-random number generator class.

    :param randoms_arr: an array intended for storing generated random numbers.
    :param generators_dim: the number of dimensions (counting from the end)
        which will use independent generators.
        For example, if ``randoms_arr`` has the shape ``(100, 200, 300)`` and
        ``generators_dim`` is ``2``, then in every sub-array ``(j, :, :)``,
        ``j = 0 .. 99``, every element will use an independent generator.
    :param sampler: a :py:class:`~reikna.cbrng.samplers.Sampler` object.
    :param seed: ``None`` for random seed, or an integer.

    .. py:classmethod:: sampler_name(randoms_arr, generators_dim, sampler_kwds=None, seed=None)

        A convenience constructor for the sampler ``sampler_name``
        from :py:mod:`~reikna.cbrng.samplers`.
        The contents of the dictionary ``sampler_kwds`` will be passed to the sampler constructor
        function (with ``bijection`` being created automatically,
        and ``dtype`` taken from ``randoms_arr``).

    .. py:method:: compiled_signature(counters:io, randoms:o)

        :param counters: the RNG "state".
            All attributes are equal to the ones of the result of :py:meth:`create_counters`.
        :param randoms: generated random numbers.
            All attributes are equal to the ones of ``randoms_arr`` from the constructor.
    """

    def __init__(self, randoms_arr, generators_dim, sampler, seed=None):

        self._sampler = sampler
        self._keygen = KeyGenerator.create(sampler.bijection, seed=seed, reserve_id_space=True)

        assert sampler.dtype == randoms_arr.dtype

        counters_size = randoms_arr.shape[-generators_dim:]

        self._generators_dim = generators_dim
        self._counters_t = Type(sampler.bijection.counter_dtype, shape=counters_size)

        Computation.__init__(self, [
            Parameter('counters', Annotation(self._counters_t, 'io')),
            Parameter('randoms', Annotation(randoms_arr, 'o'))])

    def create_counters(self):
        """
        Create a counter array for use in :py:class:`~reikna.cbrng.CBRNG`.
        """
        return numpy.zeros(self._counters_t.shape, self._counters_t.dtype)

    def _build_plan(self, plan_factory, _device_params, counters, randoms):

        plan = plan_factory()

        plan.kernel_call(
            TEMPLATE.get_def('cbrng'),
            [counters, randoms],
            global_size=helpers.product(counters.shape),
            render_kwds=dict(
                sampler=self._sampler,
                keygen=self._keygen,
                batch=helpers.product(randoms.shape[:-self._generators_dim]),
                counters_slices=[self._generators_dim],
                randoms_slices=[
                    len(randoms.shape) - self._generators_dim,
                    self._generators_dim]))

        return plan


# For some reason, closure did not work correctly.
# This class encapsulates the context and provides a classmethod for a given sampler.
class _ConvenienceCtr:

    def __init__(self, sampler_name):
        self._sampler_func = SAMPLERS[sampler_name]

    def __call__(self, cls, randoms_arr, generators_dim, sampler_kwds=None, seed=None):
        bijection = philox(64, 4)
        sampler = self._sampler_func(bijection, randoms_arr.dtype, **sampler_kwds)
        return cls(randoms_arr, generators_dim, sampler, seed=seed)


# Add convenience constructors to CBRNG
for name in SAMPLERS:
    ctr = _ConvenienceCtr(name)
    setattr(CBRNG, name, classmethod(ctr))



"""
This module contains various auxiliary functions which are used throughout the library.
"""

from __future__ import division

import functools
import itertools
import collections
import os.path
import warnings
import inspect
import funcsigs

from mako.template import Template


class Graph:

    def __init__(self, pairs=None):
        self._pairs = set()
        self._nodes = collections.defaultdict(set)
        if pairs is not None:
            self.add_edges(pairs)

    def add_edge(self, node1, node2):
        assert node1 != node2
        self._nodes[node1].add(node2)
        self._nodes[node2].add(node1)
        self._pairs.add(tuple(sorted((node1, node2))))

    def add_edges(self, pairs):
        for node1, node2 in pairs:
            self.add_edge(node1, node2)

    def add_graph(self, graph):
        for node1, node2 in graph.pairs():
            self.add_edge(node1, node2)

    def add_cluster(self, nodes):
        self.add_edges(itertools.combinations(nodes, 2))

    def remove_node(self, node):
        deps = self._nodes[node]
        for dep in deps:
            self._nodes[dep].remove(node)
            self._pairs.remove(tuple(sorted((node, dep))))
        del self._nodes[node]

    def remove_edge(self, node1, node2):
        assert node1 != node2
        self._pairs.remove(tuple(sorted((node1, node2))))

        self._nodes[node1].remove(node2)
        if len(self._nodes[node1]) == 0:
            del self._nodes[node1]

        self._nodes[node2].remove(node1)
        if len(self._nodes[node2]) == 0:
            del self._nodes[node2]

    def __getitem__(self, node):
        return self._nodes[node]

    def pairs(self):
        return self._pairs

    def translate(self, translator):
        pairs = []
        for node1, node2 in self._pairs:
            pairs.append(tuple(sorted((translator(node1), translator(node2)))))
        return Graph(pairs)


def product(seq):
    """
    Returns the product of elements in the iterable ``seq``.
    """
    return functools.reduce(lambda x1, x2: x1 * x2, seq, 1)


def make_template(template, filename=False):
    # Can't import submodules from reikna itself here, because it creates circular dependencies
    # (computation modules have template_for() calls at the root level,
    # so they get activated on import).
    kwds = dict(
        future_imports=['division'],
        strict_undefined=True,
        imports=['import numpy'])

    # Creating a template from a filename results in more comprehensible stack traces,
    # so we are taking advantage of this if possible.
    if filename:
        kwds['filename'] = template
        return Template(**kwds)
    else:
        return Template(template, **kwds)


def template_from(template):
    """
    Creates a Mako template object from a given string.
    If ``template`` already has ``render()`` method, does nothing.
    """
    if hasattr(template, 'render'):
        return template
    else:
        return make_template(template)


def extract_signature_and_value(func_or_str, default_parameters=None):
    if not inspect.isfunction(func_or_str):
        if default_parameters is None:
            parameters = []
        else:
            kind = funcsigs.Parameter.POSITIONAL_OR_KEYWORD
            parameters = [funcsigs.Parameter(name, kind=kind) for name in default_parameters]

        return funcsigs.Signature(parameters), func_or_str

    signature = funcsigs.signature(func_or_str)

    # pass mock values to extract the value
    args = [None] * len(signature.parameters)
    return signature, func_or_str(*args)


def template_def(signature, code):
    """
    Returns a ``Mako`` template with the given ``signature``.

    :param signature: a list of postitional argument names,
        or a ``Signature`` object from ``funcsigs`` module.
    :code: a body of the template.
    """
    if not isinstance(signature, funcsigs.Signature):
        # treating ``signature`` as a list of positional arguments
        # HACK: Signature or Parameter constructors are not documented.
        kind = funcsigs.Parameter.POSITIONAL_OR_KEYWORD
        signature = funcsigs.Signature([funcsigs.Parameter(name, kind=kind) for name in signature])

    template_src = "<%def name='_func" + str(signature) + "'>\n" + code + "\n</%def>"
    return template_from(template_src).get_def('_func')


def template_for(filename):
    """
    Returns the Mako template object created from the file
    which has the same name as ``filename`` and the extension ``.mako``.
    Typically used in computation modules as ``template_for(__filename__)``.
    """
    name, _ext = os.path.splitext(os.path.abspath(filename))
    return make_template(name + '.mako', filename=True)


def min_blocks(length, block):
    """
    Returns minimum number of blocks with length ``block``
    necessary to cover the array with length ``length``.
    """
    return (length - 1) // block + 1


def log2(num):
    """
    Integer-valued logarigthm with base 2.
    If ``n`` is not a power of 2, the result is rounded to the smallest number.
    """
    pos = 0
    for pow_ in [16, 8, 4, 2, 1]:
        if num >= 2 ** pow_:
            num //= (2 ** pow_)
            pos += pow_
    return pos


def bounding_power_of_2(num):
    """
    Returns closest number of the form ``2**m`` such it is greater or equal to ``n``.
    """
    return 2 ** (log2(num - 1) + 1)


def factors(num, limit=None):
    """
    Returns the list of pairs ``(factor, num/factor)`` for all factors of ``num``
    (including 1 and ``num``), sorted by ``factor``.
    If ``limit`` is set, only pairs with ``factor <= limit`` are returned.
    """
    if limit is None or limit > num:
        limit = num

    float_sqrt = num ** 0.5
    int_sqrt = int(round(float_sqrt))

    result = []

    if int_sqrt ** 2 == num:
        int_limit = int_sqrt + 1
    else:
        int_limit = int(float_sqrt) + 1

    for i in range(1, int_limit):
        div, mod = divmod(num, i)
        if mod == 0:
            result.append((i, div))

    if limit > result[-1][0]:
        if int_sqrt ** 2 == num:
            to_rev = result[:-1]
        else:
            to_rev = result

        result = result + [(div, f) for f, div in reversed(to_rev)]

    return [r for r in result if r[0] <= limit]


def wrap_in_tuple(seq_or_elem):
    """
    If ``seq_or_elem`` is a sequence, converts it to a ``tuple``,
    otherwise returns a tuple with a single element ``seq_or_elem``.
    """
    if seq_or_elem is None:
        return tuple()
    elif isinstance(seq_or_elem, str):
        return (seq_or_elem,)
    elif isinstance(seq_or_elem, collections.Iterable):
        return tuple(seq_or_elem)
    else:
        return (seq_or_elem,)


class ignore_integer_overflow():
    """
    Context manager for ignoring integer overflow in numpy operations on scalars
    (not ignored by default because of a bug in numpy).
    """

    def __init__(self):
        self.catch = warnings.catch_warnings()

    def __enter__(self):
        self.catch.__enter__()
        warnings.filterwarnings("ignore", "overflow encountered in uint_scalars")
        warnings.filterwarnings("ignore", "overflow encountered in ulong_scalars")

    def __exit__(self, *args, **kwds):
        self.catch.__exit__(*args, **kwds)



from itertools import product
import gc, re

import numpy


def get_apis(config):
    """
    Get list of APIs to test, based on command line options and their availability.
    """
    # if we import it in the header, it messes up with coverage results
    import reikna.cluda as cluda

    conf_api_id = config.option.api

    if conf_api_id == "supported":
        api_ids = cluda.supported_api_ids()
    else:
        if not cluda.supports_api(conf_api_id):
            raise Exception("Requested API " + conf_api_id + " is not supported.")
        api_ids = [conf_api_id]
    apis = [cluda.get_api(api_id) for api_id in api_ids]
    return apis, api_ids


class ThreadParams:
    """
    Encapsulates a set of parameters necessary to create a test Thread.
    """

    def __init__(self, api, pnum, dnum):
        self.api_id = api.get_id()
        self.pnum = pnum
        self.dnum = dnum

        self._api = api

        platform = api.get_platforms()[pnum]
        self._device = platform.get_devices()[dnum]

        platform_name = platform.name
        device_name = self._device.name

        self.device_params = api.DeviceParameters(self._device)

        self.device_id = "{api},{pnum},{dnum}".format(api=api.get_id(), pnum=pnum, dnum=dnum)
        self.device_full_name = platform_name + ", " + device_name

        self.id = self.device_id

        # if we import it in the header, it messes up with coverage results
        import reikna.cluda as cluda

        self.cuda = (api.get_id() == cluda.cuda_id())

    def create_thread(self):
        return self._api.Thread(self._device)

    def __key(self):
        return (self.api_id, self.pnum, self.dnum)

    def __eq__(x, y):
        return x.__key() == y.__key()

    def __hash__(self):
        return hash(self.__key())

    def __str__(self):
        return self.id


class ThreadManager:
    """
    Caches Thread instances and handles push/pop mechanics of CUDA Threads properly.
    """

    def __init__(self):
        self._threads = {}

    def _create(self, thread_params):

        thread = thread_params.create_thread()

        # The context is in the stack right after creation, need to pop it.
        if thread_params.cuda:
            thread._cuda_pop()

        self._threads[thread_params] = thread

    def get(self, thread_params):

        if thread_params not in self._threads:
            self._create(thread_params)

        thread = self._threads[thread_params]
        if thread_params.cuda:
            thread._cuda_push()

        return thread

    def release(self, thread_params):
        thread = self._threads[thread_params]
        if thread_params.cuda:
            thread._cuda_pop()

_thread_manager = ThreadManager()


def get_threads(config):
    """
    Create a list of thread creators, based on command line options and their availability.
    """
    # if we import it in the header, it messes up with coverage results
    from reikna.cluda import find_devices

    apis, _ = get_apis(config)

    tps = []
    for api in apis:
        devices = find_devices(
            api,
            include_devices=config.option.device_include_mask,
            exclude_devices=config.option.device_exclude_mask,
            include_platforms=config.option.platform_include_mask,
            exclude_platforms=config.option.platform_exclude_mask,
            include_duplicate_devices=config.option.include_duplicate_devices)

        for pnum in sorted(devices.keys()):
            dnums = sorted(devices[pnum])
            for dnum in dnums:
                tps.append(ThreadParams(api, pnum, dnum))

    return tps, [str(tp) for tp in tps]


def pair_thread_with_doubles(metafunc, tp):
    d = metafunc.config.option.double
    ds = lambda dv: 'dp' if dv else 'sp'

    vals = []
    ids = []
    if d == "supported":
        for dv in [False, True]:
            if not dv or tp.device_params.supports_dtype(numpy.float64):
                vals.append((dv,))
                ids.append(ds(dv))
    else:
        dv = d == 'yes'
        vals.append((dv,))
        ids.append(ds(dv))

    return [tp] * len(vals), vals, ids


def get_thread_tuples(metafunc, get_remainders):
    tps, tp_ids = get_threads(metafunc.config)
    tuples = []
    ids = []
    for tp, tp_id in zip(tps, tp_ids):
        new_tps, remainders, rem_ids = get_remainders(metafunc, tp)
        for new_tp, remainder, rem_id in zip(new_tps, remainders, rem_ids):
            tuples.append((new_tp,) + remainder if isinstance(remainder, tuple) else (remainder,))
            ids.append(tp_id + "," + rem_id)

    # For tuples of 1 element (i.e. the thread itself), just use this element as a parameter
    tuples = [t[0] if len(t) == 1 else t for t in tuples]

    return tuples, ids


def create_thread_in_tuple(request):
    """
    Instantiate thread from the tuple of test parameters and create a corresponding finalizer.
    """
    params = request.param
    if isinstance(params, tuple):
        tp = params[0]
        remainder = tuple(params[1:])
    else:
        tp = params
        remainder = tuple()

    thr = _thread_manager.get(tp)

    def finalizer():
        # Py.Test holds the reference to the created funcarg/fixture,
        # which interferes with ``__del__`` functionality.
        # This method forcefully frees critical resources explicitly
        # (rendering the object unusable).
        _thread_manager.release(tp)
        # just in case there is some stuff left
        gc.collect()

    request.addfinalizer(finalizer)

    if isinstance(params, tuple):
        return (thr,) + remainder
    else:
        return thr


def parametrize_thread_tuple(metafunc, name, get_remainders):
    tuples, tuple_ids = get_thread_tuples(metafunc, get_remainders)
    metafunc.parametrize(name, tuples, ids=tuple_ids, indirect=True)



import reikna.helpers as helpers
from reikna.core import Computation, Parameter, Annotation, Type
import reikna.cluda.dtypes as dtypes
from reikna.cluda import OutOfResourcesError
from reikna.cluda import functions

TEMPLATE = helpers.template_for(__file__)


class MatrixMul(Computation):
    """
    Bases: :py:class:`~reikna.core.Computation`

    Multiplies two matrices using last two dimensions and batching over remaining dimensions.
    For batching to work, the products of remaining dimensions should be equal
    (then the multiplication will be performed piecewise), or one of them should equal 1
    (then the multiplication will be batched over the remaining dimensions of the other matrix).

    :param a_arr: an array-like defining the first argument.
    :param b_arr: an array-like defining the second argument.
    :param out_arr: an array-like definign the output; if not given, both shape and dtype
        will be derived from ``a_arr`` and ``b_arr``.
    :param block_width_override: if provided, it will used as a block size of
        the multiplication kernel.

    .. py:method:: compiled_signature(output:o, matrix_a:i, matrix_b:i)

        :param output: the output of matrix multiplication.
        :param matrix_a: the first argument.
        :param matrix_b: the second argument.
    """

    def __init__(self, a_arr, b_arr, out_arr=None, block_width_override=None):

        if len(a_arr.shape) == 1:
            a_arr = Type(a_arr.dtype, shape=(1,) + a_arr.shape)

        if len(b_arr.shape) == 1:
            b_arr = Type(b_arr.dtype, shape=b_arr.shape + (1,))

        if out_arr is None:
            out_dtype = dtypes.result_type(a_arr.dtype, b_arr.dtype)

            a_batch = helpers.product(a_arr.shape[:-2])
            batch_len = max(len(a_arr.shape[:-2]), len(b_arr.shape[:-2]))
            batch_shape = b_arr.shape[:-2] if a_batch == 1 else a_arr.shape[:-2]
            batch_shape = (1,) * (batch_len - len(batch_shape)) + batch_shape

            out_shape = batch_shape + (a_arr.shape[-2], b_arr.shape[-1])

            out_arr = Type(out_dtype, shape=out_shape)

        Computation.__init__(self, [
            Parameter('output', Annotation(out_arr, 'o')),
            Parameter('matrix_a', Annotation(a_arr, 'i')),
            Parameter('matrix_b', Annotation(b_arr, 'i'))])

        self._block_width_override = block_width_override

    def _build_plan(self, plan_factory, device_params, output, matrix_a, matrix_b):
        bwo = self._block_width_override

        if bwo is not None:
            block_widths = [bwo]
        else:
            nbanks = device_params.local_mem_banks
            block_widths = [2 ** n for n in range(helpers.log2(nbanks), -1, -1)]

        a_height = matrix_a.shape[-2]
        a_width = matrix_a.shape[-1]
        b_width = matrix_b.shape[-1]
        a_batch = helpers.product(matrix_a.shape[:-2])
        b_batch = helpers.product(matrix_b.shape[:-2])
        batch = max(a_batch, b_batch)

        for block_width in block_widths:

            plan = plan_factory()

            if block_width ** 2 > device_params.max_work_group_size:
                continue

            num_steps = helpers.min_blocks(a_width, block_width)
            blocks_per_matrix = helpers.min_blocks(a_height, block_width)
            grid_width = helpers.min_blocks(b_width, block_width)

            render_kwds = dict(
                batched_a=(a_batch != 1),
                batched_b=(b_batch != 1),
                a_height=a_height,
                b_width=b_width,
                a_width=a_width,
                num_steps=num_steps,
                a_slices=(len(matrix_a.shape) - 2, 1, 1),
                b_slices=(len(matrix_b.shape) - 2, 1, 1),
                output_slices=(len(output.shape) - 2, 1, 1),
                block_width=block_width,
                mul=functions.mul(matrix_a.dtype, matrix_b.dtype, out_dtype=output.dtype))

            try:
                plan.kernel_call(
                    TEMPLATE.get_def('matrixmul'),
                    [output, matrix_a, matrix_b],
                    global_size=(
                        batch,
                        blocks_per_matrix * block_width,
                        grid_width * block_width),
                    local_size=(1, block_width, block_width),
                    render_kwds=render_kwds)
            except OutOfResourcesError:
                continue

            return plan

        raise ValueError("Could not find suitable call parameters for the kernel")



import numpy
import pytest

from helpers import *
from test_core.dummy import *


def test_dummy(thr):

    N = 200
    coeff = 2
    A = get_test_array((N, N), numpy.complex64)
    B = get_test_array(N, numpy.complex64)

    A_dev = thr.to_device(A)
    B_dev = thr.to_device(B)
    C_dev = thr.empty_like(A_dev)
    D_dev = thr.empty_like(B_dev)

    d = Dummy(A_dev, B_dev, numpy.float32).compile(thr)
    d(C_dev, D_dev, A_dev, B_dev, coeff)
    C = C_dev.get()
    D = D_dev.get()

    C_ref, D_ref = mock_dummy(A, B, coeff)
    assert diff_is_negligible(C, C_ref)
    assert diff_is_negligible(D, D_ref)


def test_nested_dummy(thr):

    N = 2000
    coeff = 2
    second_coeff = 3
    A = get_test_array((N, N), numpy.complex64)
    B = get_test_array(N, numpy.complex64)

    A_dev = thr.to_device(A)
    B_dev = thr.to_device(B)
    C_dev = thr.empty_like(A_dev)
    D_dev = thr.empty_like(B_dev)

    d = DummyNested(A_dev, B_dev, numpy.float32, second_coeff).compile(thr)
    d(C_dev, D_dev, A_dev, B_dev, coeff)
    C = C_dev.get()
    D = D_dev.get()

    C_ref, D_ref = mock_dummy_nested(A, B, coeff, second_coeff)

    assert diff_is_negligible(C, C_ref)
    assert diff_is_negligible(D, D_ref)


def test_incorrect_parameter_name():
    """
    Tests that setting incorrect parameter name
    in the Computation constructor raises an exception.
    """

    N = 200
    A = get_test_array((N, N), numpy.complex64)
    B = get_test_array(N, numpy.complex64)

    with pytest.raises(ValueError):
        d = Dummy(A, B, numpy.float32, test_incorrect_parameter_name=True)


def test_untyped_scalar(some_thr):
    """
    Tests that passing an untyped (i.e. not a ndarray with shape==tuple())
    scalar as an argument to a kernel while creating a plan raises an exception.
    """

    N = 200
    A = get_test_array((N, N), numpy.complex64)
    B = get_test_array(N, numpy.complex64)

    d = Dummy(A, B, numpy.float32, test_untyped_scalar=True)
    with pytest.raises(TypeError):
        dc = d.compile(some_thr)


def test_kernel_adhoc_array(some_thr):
    """
    Tests that passing an array as an ad hoc argument to a kernel
    while creating a plan raises an exception.
    """

    N = 200
    A = get_test_array((N, N), numpy.complex64)
    B = get_test_array(N, numpy.complex64)

    d = Dummy(A, B, numpy.float32, test_kernel_adhoc_array=True)
    with pytest.raises(ValueError):
        dc = d.compile(some_thr)


def test_computation_adhoc_array(some_thr):
    """
    Tests that passing an array as an ad hoc argument to a computation
    while creating a plan raises an exception.
    """

    N = 200
    A = get_test_array((N, N), numpy.complex64)
    B = get_test_array(N, numpy.complex64)

    d = DummyNested(A, B, numpy.float32, 3, test_computation_adhoc_array=True)
    with pytest.raises(ValueError):
        dc = d.compile(some_thr)


def test_computation_incorrect_role(some_thr):
    """
    Tests that passing an array which does not support a required role
    (e.g. a role=='i' for role=='o' parameter) as an argument to a computation
    while creating a plan raises an exception.
    """

    N = 200
    A = get_test_array((N, N), numpy.complex64)
    B = get_test_array(N, numpy.complex64)

    d = DummyNested(A, B, numpy.float32, 3, test_computation_incorrect_role=True)
    with pytest.raises(TypeError):
        dc = d.compile(some_thr)


def test_computation_incorrect_type(some_thr):
    """
    Tests that passing an argument with an incorrect type as an argument to a computation
    while creating a plan raises an exception.
    """

    N = 200
    A = get_test_array((N, N), numpy.complex64)
    B = get_test_array(N, numpy.complex64)

    d = DummyNested(A, B, numpy.float32, 3, test_computation_incorrect_type=True)
    with pytest.raises(TypeError):
        dc = d.compile(some_thr)


def test_same_arguments(some_thr):
    """
    Tests passing the same KernelArgument for two parameters of a computation or a kernel.
    """

    N = 200
    coeff1 = 3
    coeff2 = 4
    C = get_test_array((N, N), numpy.complex64)
    D = get_test_array((N, N), numpy.complex64)
    C_dev = some_thr.to_device(C)
    D_dev = some_thr.to_device(D)

    d = DummyAdvanced(C, numpy.float32)
    dc = d.compile(some_thr)

    dc(C_dev, D_dev, coeff1, coeff2)
    C_ref, D_ref = mock_dummy_advanced(C, D, coeff1, coeff2)

    C = C_dev.get()
    D = D_dev.get()

    assert diff_is_negligible(C, C_ref)
    assert diff_is_negligible(D, D_ref)



import numpy

import reikna.helpers as helpers
import reikna.cluda.dtypes as dtypes
from reikna.cluda import Module

TEMPLATE = helpers.template_for(__file__)


class Sampler:
    """
    Contains a random distribution sampler module and accompanying metadata.
    Supports ``__process_modules__`` protocol.

    .. py:attribute:: deterministic

        If ``True``, every sampled random number consumes the same amount of counters.

    .. py:attribute:: randoms_per_call

        How many random numbers one call to ``sample`` creates.

    .. py:attribute:: dtype

        The data type of one random value produced by the sampler.

    .. py:attribute:: module

        The module containing the distribution sampling function.
        It provides the C functions below.

    .. c:macro:: RANDOMS_PER_CALL

        Contains the value of :py:attr:`randoms_per_call`.

    .. c:type:: Value

        Contains the type corresponding to :py:attr:`dtype`.

    .. c:type:: Result

        Describes the sampling result.

        .. c:member:: value v[RANDOMS_PER_CALL]

    .. c:type:: Result sample(State *state)

        Performs the sampling, updating the state.
    """
    def __init__(self, bijection, module, dtype, randoms_per_call=1, deterministic=False):
        """__init__()""" # hide the signature from Sphinx
        self.randoms_per_call = randoms_per_call
        self.dtype = dtypes.normalize_type(dtype)
        self.deterministic = deterministic
        self.bijection = bijection
        self.module = module

    def __process_modules__(self, process):
        return Sampler(
            process(self.bijection), process(self.module), self.dtype,
            randoms_per_call=self.randoms_per_call,
            deterministic=self.deterministic)


def uniform_integer(bijection, dtype, low, high=None):
    """
    Generates uniformly distributed integer numbers in the interval ``[low, high)``.
    If ``high`` is ``None``, the interval is ``[0, low)``.
    Supported dtypes: any numpy integers.
    If the size of the interval is a power of 2, a fixed number of counters
    is used in each thread.
    Returns a :py:class:`~reikna.cbrng.samplers.Sampler` object.
    """

    if high is None:
        low, high = 0, low + 1
    else:
        assert low < high - 1

    dtype = dtypes.normalize_type(dtype)
    ctype = dtypes.ctype(dtype)

    if dtype.kind == 'i':
        assert low >= -2 ** (dtype.itemsize * 8 - 1)
        assert high < 2 ** (dtype.itemsize * 8 - 1)
    else:
        assert low >= 0
        assert high < 2 ** (dtype.itemsize * 8)

    num = high - low
    if num <= 2 ** 32:
        raw_dtype = numpy.dtype('uint32')
    else:
        raw_dtype = numpy.dtype('uint64')

    raw_func = bijection.raw_functions[raw_dtype]
    max_num = 2 ** (raw_dtype.itemsize * 8)

    raw_ctype = dtypes.ctype(dtypes.normalize_type(raw_dtype))

    module = Module(
        TEMPLATE.get_def("uniform_integer"),
        render_kwds=dict(
            bijection=bijection,
            dtype=dtype, ctype=ctype,
            raw_ctype=raw_ctype, raw_func=raw_func,
            max_num=max_num, num=num, low=low))

    return Sampler(bijection, module, dtype, deterministic=(max_num % num == 0))


def uniform_float(bijection, dtype, low=0, high=1):
    """
    Generates uniformly distributed floating-points numbers in the interval ``[low, high)``.
    Supported dtypes: ``float(32/64)``.
    A fixed number of counters is used in each thread.
    Returns a :py:class:`~reikna.cbrng.samplers.Sampler` object.
    """
    assert low < high

    ctype = dtypes.ctype(dtype)

    bitness = 64 if dtypes.is_double(dtype) else 32
    raw_func = 'get_raw_uint' + str(bitness)
    raw_max = dtypes.c_constant(2 ** bitness, dtype)

    size = dtypes.c_constant(high - low, dtype)
    low = dtypes.c_constant(low, dtype)

    module = Module(
        TEMPLATE.get_def("uniform_float"),
        render_kwds=dict(
            bijection=bijection, ctype=ctype,
            raw_func=raw_func, raw_max=raw_max, size=size, low=low))

    return Sampler(bijection, module, dtype, deterministic=True)


def normal_bm(bijection, dtype, mean=0, std=1):
    """
    Generates normally distributed random numbers with the mean ``mean`` and
    the standard deviation ``std`` using Box-Muller transform.
    Supported dtypes: ``float(32/64)``.
    Produces two random numbers per call.
    Returns a :py:class:`~reikna.cbrng.samplers.Sampler` object.
    """

    ctype = dtypes.ctype(dtype)
    uf = uniform_float(bijection, dtype, low=0, high=1)

    module = Module(
        TEMPLATE.get_def("normal_bm"),
        render_kwds=dict(
            dtype=dtype, ctype=ctype, bijection=bijection,
            mean=dtypes.c_constant(mean, dtype),
            std=dtypes.c_constant(std, dtype),
            uf=uf))

    return Sampler(
        bijection, module, dtype,
        deterministic=uf.deterministic, randoms_per_call=2)


def gamma(bijection, dtype, shape=1, scale=1):
    """
    Generates random numbers from the gamma distribution

    .. math::
      P(x) = x^{k-1} \\frac{e^{-x/\\theta}}{\\theta^k \\Gamma(k)},

    where :math:`k` is ``shape``, and :math:`\\theta` is ``scale``.
    Supported dtypes: ``float(32/64)``.
    Returns a :py:class:`~reikna.cbrng.samplers.Sampler` object.
    """

    ctype = dtypes.ctype(dtype)
    uf = uniform_float(bijection, dtype, low=0, high=1)
    nbm = normal_bm(bijection, dtype, mean=0, std=1)

    module = Module(
        TEMPLATE.get_def("gamma"),
        render_kwds=dict(
            dtype=dtype, ctype=ctype, bijection=bijection,
            shape=shape, scale=dtypes.c_constant(scale, dtype),
            uf=uf, nbm=nbm))

    return Sampler(bijection, module, dtype)


# List of samplers that can be used as convenience constructors in CBRNG class
SAMPLERS = {
    'uniform_integer': uniform_integer,
    'uniform_float': uniform_float,
    'normal_bm': normal_bm,
    'gamma': gamma}



import itertools

import pytest

import reikna.cluda as cluda
import reikna.cluda.dtypes as dtypes
import reikna.cluda.functions as functions
from reikna.cluda import tempalloc
from reikna.helpers import product

from helpers import *
from pytest_threadgen import parametrize_thread_tuple, create_thread_in_tuple


TEST_DTYPES = [
    numpy.int8, numpy.int16, numpy.int32, numpy.int64,
    numpy.uint8, numpy.uint16, numpy.uint32, numpy.uint64,
    numpy.float32, numpy.float64,
    numpy.complex64, numpy.complex128]


pytest_funcarg__thr_and_global_size = create_thread_in_tuple


def pair_thread_with_gs(metafunc, tp):
    global_sizes = [
        (100,), (2000,), (1153,),
        (10, 10), (150, 250), (137, 547),
        (7, 11, 13), (50, 100, 100), (53, 101, 101)]

    rem_ids = []
    vals = []

    for gs in global_sizes:

        # If the thread will not support these limits, skip
        mgs = tp.device_params.max_num_groups
        if len(gs) > len(mgs) or (len(mgs) > 2 and len(gs) > 2 and mgs[2] < gs[2]):
            continue

        rem_ids.append(str(gs))
        vals.append((gs,))

    return [tp] * len(vals), vals, rem_ids


def pytest_generate_tests(metafunc):
    if 'thr_and_global_size' in metafunc.funcargnames:
        parametrize_thread_tuple(metafunc, 'thr_and_global_size', pair_thread_with_gs)


def simple_thread_test(thr):
    shape = (1000,)
    dtype = numpy.float32

    a = get_test_array(shape, dtype)
    a_dev = thr.to_device(a)
    a_back = thr.from_device(a_dev)

    assert diff_is_negligible(a, a_back)


def test_create_new_thread(cluda_api):
    thr = cluda_api.Thread.create()
    simple_thread_test(thr)


def test_transfers(thr):
    a = get_test_array(1024, numpy.float32)

    def to_device1(x):
        return thr.to_device(x)
    def to_device2(x):
        y = thr.empty_like(x)
        thr.to_device(x, dest=y)
        return y
    def from_device1(x):
        return x.get()
    def from_device2(x):
        return thr.from_device(x)
    def from_device3(x):
        y = numpy.empty(x.shape, x.dtype)
        thr.from_device(x, dest=y)
        return y
    def from_device4(x):
        y = thr.from_device(x, async=True)
        thr.synchronize()
        return y
    def from_device5(x):
        y = numpy.empty(x.shape, x.dtype)
        thr.from_device(x, dest=y, async=True)
        thr.synchronize()
        return y

    to_device = (to_device1, to_device2)
    from_device = (from_device1, from_device2, from_device3, from_device4, from_device5)

    for to_d, from_d in itertools.product(to_device, from_device):
        a_device = to_d(a)
        a_copy = thr.copy_array(a_device)
        a_back = from_d(a_copy)
        assert diff_is_negligible(a, a_back)


@pytest.mark.parametrize(
    "dtype", TEST_DTYPES,
    ids=[dtypes.normalize_type(dtype).name for dtype in TEST_DTYPES])
def test_dtype_support(thr, dtype):
    # Test passes if either thread correctly reports that it does not support given dtype,
    # or it successfully compiles kernel that operates with this dtype.

    N = 256

    if not thr.device_params.supports_dtype(dtype):
        pytest.skip()

    mul = functions.mul(dtype, dtype)
    div = functions.div(dtype, dtype)
    program = thr.compile(
    """
    KERNEL void test(
        GLOBAL_MEM ${ctype} *dest, GLOBAL_MEM ${ctype} *a, GLOBAL_MEM ${ctype} *b)
    {
      const SIZE_T i = get_global_id(0);
      ${ctype} temp = ${mul}(a[i], b[i]);
      dest[i] = ${div}(temp, b[i]);
    }
    """, render_kwds=dict(ctype=dtypes.ctype(dtype), dtype=dtype, mul=mul, div=div))

    test = program.test

    # we need results to fit even in unsigned char
    a = get_test_array(N, dtype, high=8)
    b = get_test_array(N, dtype, no_zeros=True, high=8)

    a_dev = thr.to_device(a)
    b_dev = thr.to_device(b)
    dest_dev = thr.empty_like(a_dev)
    test(dest_dev, a_dev, b_dev, global_size=N)
    assert diff_is_negligible(thr.from_device(dest_dev), a)


def test_find_local_size(thr_and_global_size):
    """
    Check that if None is passed as local_size, kernel can find some local_size to run with
    (not necessarily optimal).
    """

    thr, global_size = thr_and_global_size

    program = thr.compile(
    """
    KERNEL void test(GLOBAL_MEM int *dest)
    {
        const SIZE_T i = get_global_id(0) +
            get_global_id(1) * get_global_size(0) +
            get_global_id(2) * get_global_size(1) * get_global_size(0);
        dest[i] = i;
    }
    """)
    test = program.test
    dest_dev = thr.array(global_size, numpy.int32)
    test(dest_dev, global_size=global_size)

    assert diff_is_negligible(dest_dev.get().ravel(),
        numpy.arange(product(global_size)).astype(numpy.int32))


@pytest.mark.parametrize(
    'tempalloc_cls',
    [tempalloc.TrivialManager, tempalloc.ZeroOffsetManager],
    ids=['trivial', 'zero_offset'])
@pytest.mark.parametrize('pack', [False, True], ids=['no_pack', 'pack'])
def test_tempalloc(cluda_api, tempalloc_cls, pack):

    shape = (10000,)
    dtype = numpy.int32
    thr = cluda_api.Thread.create(temp_alloc=dict(
        cls=tempalloc_cls, pack_on_alloc=False))

    # Dependency graph for the test
    dependencies = dict(
        _temp0=[],
        _temp1=['_temp9', '_temp8', '_temp3', '_temp5', '_temp4', '_temp7', '_temp6', 'input'],
        _temp10=['output', '_temp7'],
        _temp11=['_temp7'],
        _temp2=['input'],
        _temp3=['_temp1', 'input'],
        _temp4=['_temp9', '_temp8', '_temp1', '_temp7', '_temp6'],
        _temp5=['_temp1'],
        _temp6=['_temp1', '_temp4'],
        _temp7=['_temp9', '_temp1', '_temp4', 'output', '_temp11', '_temp10'],
        _temp8=['_temp1', '_temp4'],
        _temp9=['_temp1', '_temp4', '_temp7'],
        input=['_temp1', '_temp3', '_temp2'],
        output=['_temp10', '_temp7'])

    program = thr.compile(
    """
    KERNEL void fill(GLOBAL_MEM ${ctype} *dest, ${ctype} val)
    {
      const SIZE_T i = get_global_id(0);
      dest[i] = val;
    }

    KERNEL void transfer(GLOBAL_MEM ${ctype} *dest, GLOBAL_MEM ${ctype} *src)
    {
      const SIZE_T i = get_global_id(0);
      dest[i] = src[i];
    }
    """, render_kwds=dict(ctype=dtypes.ctype(dtype)))
    fill = program.fill
    transfer = program.transfer

    arrays = {}
    transfer_dest = thr.array(shape, dtype)

    # Allocate temporary arrays with dependencies
    for name in sorted(dependencies.keys()):
        deps = dependencies[name]
        arr_deps = [arrays[d] for d in deps if d in arrays]
        arrays[name] = thr.temp_array(shape, dtype, dependencies=arr_deps)
        fill(arrays[name], dtype(0), global_size=shape)

    if pack:
        thr.temp_alloc.pack()

    # Fill arrays with zeros
    for name in sorted(dependencies.keys()):
        deps = dependencies[name]
        arr_deps = [arrays[d] for d in deps if d in arrays]
        arrays[name] = thr.temp_array(shape, dtype, dependencies=arr_deps)
        fill(arrays[name], dtype(0), global_size=shape)

    for i, name in enumerate(sorted(dependencies.keys())):
        val = dtype(i + 1)
        fill(arrays[name], val, global_size=shape)
        for dep in dependencies[name]:
            # CUDA does not support get() for GPUArray with custom buffers,
            # So we need to transfer the data to a normal array first.
            transfer(transfer_dest, arrays[dep], global_size=shape)
            assert (transfer_dest.get() != val).all()



import re


def _name_matches_masks(name, includes, excludes):
    if len(includes) > 0:
        for include in includes:
            if re.search(include, name):
                break
        else:
            return False

    if len(excludes) > 0:
        for exclude in excludes:
            if re.search(exclude, name):
                return False

    return True


def find_devices(
        api,
        include_devices=None, exclude_devices=None,
        include_platforms=None, exclude_platforms=None,
        include_duplicate_devices=True,
        include_pure_only=False):
    """
    Find platforms and devices meeting certain criteria.

    :param api: a CLUDA API object.
    :param include_devices: a list of masks for a device name
        which will be used to pick devices to include in the result.
    :param exclude_devices: a list of masks for a device name
        which will be used to pick devices to exclude from the result.
    :param include_platforms: a list of masks for a platform name
        which will be used to pick platforms to include in the result.
    :param exclude_platforms: a list of masks for a platform name
        which will be used to pick platforms to exclude in the result.
    :param include_duplicate_devices: if ``False``, will only include a single device
        from the several with the same name available on a platform.
    :param include_pure_only: if ``True``, will include devices with maximum group size equal to 1.
    :returns: a dictionary with found platform numbers as keys,
        and lists of device numbers as values.
    """

    if include_devices is None:
        include_devices = []
    if exclude_devices is None:
        exclude_devices = []
    if include_platforms is None:
        include_platforms = []
    if exclude_platforms is None:
        exclude_platforms = []

    devices = {}

    for pnum, platform in enumerate(api.get_platforms()):

        seen_devices = set()

        if not _name_matches_masks(platform.name, include_platforms, exclude_platforms):
            continue

        for dnum, device in enumerate(platform.get_devices()):
            if not _name_matches_masks(device.name, include_devices, exclude_devices):
                continue

            if not include_duplicate_devices and device.name in seen_devices:
                continue

            params = api.DeviceParameters(device)
            if not include_pure_only and params.max_work_group_size == 1:
                continue

            seen_devices.add(device.name)

            if pnum not in devices:
                devices[pnum] = []

            devices[pnum].append(dnum)

    return devices



import numpy
from reikna.cluda.api_discovery import cuda_id, ocl_id


_DTYPE_TO_BUILTIN_CTYPE = {}
_DTYPE_TO_CTYPE_MODULE = {}


def is_complex(dtype):
    """
    Returns ``True`` if ``dtype`` is complex.
    """
    dtype = normalize_type(dtype)
    return dtype.kind == 'c'

def is_double(dtype):
    """
    Returns ``True`` if ``dtype`` is double precision floating point.
    """
    dtype = normalize_type(dtype)
    return dtype.name in ['float64', 'complex128']

def is_integer(dtype):
    """
    Returns ``True`` if ``dtype`` is an integer.
    """
    dtype = normalize_type(dtype)
    return dtype.kind in ('i', 'u')

def is_real(dtype):
    """
    Returns ``True`` if ``dtype`` is a real.
    """
    dtype = normalize_type(dtype)
    return dtype.kind == 'f'

def _promote_dtype(dtype):
    # not all numpy datatypes are supported by GPU, so we may need to promote
    dtype = normalize_type(dtype)
    if dtype.kind == 'i' and dtype.itemsize < 4:
        dtype = numpy.int32
    elif dtype.kind == 'f' and dtype.itemsize < 4:
        dtype = numpy.float32
    elif dtype.kind == 'c' and dtype.itemsize < 8:
        dtype = numpy.complex64
    return normalize_type(dtype)

def result_type(*dtypes):
    """
    Wrapper for ``numpy.result_type``
    which takes into account types supported by GPUs.
    """
    return _promote_dtype(numpy.result_type(*dtypes))

def min_scalar_type(val):
    """
    Wrapper for ``numpy.min_scalar_dtype``
    which takes into account types supported by GPUs.
    """
    return _promote_dtype(numpy.min_scalar_type(val))

def detect_type(val):
    """
    Find out the data type of ``val``.
    """
    if hasattr(val, 'dtype'):
        return _promote_dtype(val.dtype)
    else:
        return min_scalar_type(val)

def normalize_type(dtype):
    """
    Function for wrapping all dtypes coming from the user.
    ``numpy`` uses two different classes to represent dtypes,
    and one of them does not have some important attributes.
    """
    return numpy.dtype(dtype)

def normalize_types(dtypes):
    """
    Same as :py:func:`normalize_type`, but operates on a list of dtypes.
    """
    return [normalize_type(dtype) for dtype in dtypes]

def complex_for(dtype):
    """
    Returns complex dtype corresponding to given floating point ``dtype``.
    """
    dtype = normalize_type(dtype)
    return numpy.dtype(dict(float32='complex64', float64='complex128')[dtype.name])

def real_for(dtype):
    """
    Returns floating point dtype corresponding to given complex ``dtype``.
    """
    dtype = normalize_type(dtype)
    return numpy.dtype(dict(complex64='float32', complex128='float64')[dtype.name])

def complex_ctr(dtype):
    """
    Returns name of the constructor for the given ``dtype``.
    """
    return 'COMPLEX_CTR(' + ctype(dtype) + ')'

def zero_ctr(dtype):
    """
    Returns the string with constructed zero value for the given ``dtype``.
    """
    if is_complex(dtype):
        return complex_ctr(dtype) + '(0, 0)'
    else:
        return '0'

def cast(dtype):
    """
    Returns function that takes one argument and casts it to ``dtype``.
    """
    return numpy.cast[dtype]

def c_constant(val, dtype=None):
    """
    Returns a C-style numerical constant.
    """
    if dtype is None:
        dtype = detect_type(val)
    else:
        dtype = normalize_type(dtype)
    val = numpy.cast[dtype](val)

    if is_complex(dtype):
        return "COMPLEX_CTR(" + ctype(dtype) + ")(" + \
            c_constant(val.real) + ", " + c_constant(val.imag) + ")"
    elif is_integer(dtype):
        return str(val) + ("L" if dtype.itemsize > 4 else "")
    else:
        return repr(float(val)) + ("f" if dtype.itemsize <= 4 else "")

def _register_dtype(dtype, ctype_str):
    dtype = normalize_type(dtype)
    _DTYPE_TO_BUILTIN_CTYPE[dtype] = ctype_str

# Taken from compyte.dtypes
def _fill_dtype_registry(respect_windows=True):

    import sys
    import platform

    _register_dtype(numpy.bool, "bool")
    _register_dtype(numpy.int8, "char")
    _register_dtype(numpy.uint8, "unsigned char")
    _register_dtype(numpy.int16, "short")
    _register_dtype(numpy.uint16, "unsigned short")
    _register_dtype(numpy.int32, "int")
    _register_dtype(numpy.uint32, "unsigned int")

    # recommended by Python docs
    is_64bits = sys.maxsize > 2 ** 32

    if is_64bits:
        if platform.system == 'Windows' and respect_windows:
            i64_name = "long long"
        else:
            i64_name = "long"

        _register_dtype(numpy.int64, i64_name)
        _register_dtype(numpy.uint64, "unsigned %s" % i64_name)

        # http://projects.scipy.org/numpy/ticket/2017
        _register_dtype(numpy.uintp, "unsigned %s" % i64_name)
    else:
        _register_dtype(numpy.uintp, "unsigned")

    _register_dtype(numpy.float32, "float")
    _register_dtype(numpy.float64, "double")
    _register_dtype(numpy.complex64, "float2")
    _register_dtype(numpy.complex128, "double2")

_fill_dtype_registry()


def adjust_alignment(thr, dtype):
    """
    Returns a new data type object with the same fields as in ``dtype``
    and the field offsets set by the compiler in ``thr``.
    All existing non-standard offsets in ``dtype`` are ignored.
    """

    if dtype.names is None:
        return dtype

    adjusted_dtypes = [
        adjust_alignment(thr, dtype.fields[name][0])
        for name in dtype.names]

    new_dtype = numpy.dtype(dict(
        names=dtype.names,
        formats=adjusted_dtypes))

    struct = ctype_module(new_dtype)

    program = thr.compile(
    """
    #define my_offsetof(type, field) ((size_t)(&((type *)0)->field))

    KERNEL void test(GLOBAL_MEM int *dest)
    {
      %for i, name in enumerate(dtype.names):
      dest[${i}] = my_offsetof(${struct}, ${name});
      %endfor
      dest[${len(dtype.names)}] = sizeof(${struct});
    }
    """, render_kwds=dict(
        struct=struct,
        dtype=new_dtype))

    offsets = thr.array(len(dtype.names) + 1, numpy.int32)
    test = program.test
    test(offsets, global_size=1)
    offsets = offsets.get()

    # Casting to Python ints, becase numpy ints as dtype offsets make it unhashable.
    offsets = [int(offset) for offset in offsets]

    return numpy.dtype(dict(
        names=dtype.names,
        formats=adjusted_dtypes,
        offsets=offsets[:-1],
        itemsize=offsets[-1]))


def ctype(dtype):
    """
    For a built-in C type, returns a string with the name of the type.
    """
    return _DTYPE_TO_BUILTIN_CTYPE[normalize_type(dtype)]


def _alignment_str(alignment):
    if alignment is not None:
        return "ALIGN(" + str(alignment) + ")"
    else:
        return ""


def _get_struct_module(dtype):
    """
    Builds and returns a module with the C type definition for a given ``dtype``,
    possibly using modules for nested structures.
    """
    if dtype.names is None:
        raise ValueError("dtype must be a struct type")

    if len(dtype.shape) > 0:
        raise ValueError("The data type cannot be an array")

    lines = ["typedef struct {"]
    kwds = {}
    for i, name in enumerate(dtype.names):
        elem_dtype, elem_offset = dtype.fields[name]

        if len(elem_dtype.shape) == 0:
            base_elem_dtype = elem_dtype
            elem_dtype_shape = tuple()
        else:
            base_elem_dtype = elem_dtype.base
            elem_dtype_shape = elem_dtype.shape

        if i == len(dtype.names) - 1:
            # If it is the last field of the struct, its alignment does not matter ---
            # the enompassing struct's one will override it anyway.
            alignment = None
        else:
            alignment = dtype.fields[dtype.names[i+1]][1] - elem_offset
            if alignment <= elem_dtype.itemsize:
                alignment = None

        if len(elem_dtype_shape) == 0:
            array_suffix = ""
        else:
            array_suffix = "".join("[" + str(d) + "]" for d in elem_dtype_shape)

        typename_var = "typename" + str(i)
        lines.append(
            "    ${" + typename_var + "} " + _alignment_str(alignment) + " " +
            name + array_suffix + ";")
        kwds[typename_var] = ctype_module(base_elem_dtype)

    expected_itemsize = sum(dt.itemsize for dt, _ in dtype.fields.values())
    if dtype.itemsize > expected_itemsize:
        alignment = dtype.itemsize
    else:
        alignment = None

    lines.append("} " + _alignment_str(alignment) + " ${prefix};")

    # Root level import creates an import loop.
    from reikna.cluda.kernel import Module

    return Module.create("\n".join(lines), render_kwds=kwds)


def ctype_module(dtype):
    """
    For a struct type, returns a :py:class:`~reikna.cluda.Module` object
    with the ``typedef`` of a struct corresponding to the given ``dtype``
    (with its name set to the module prefix);
    falls back to :py:func:`~reikna.cluda.dtypes.ctype` otherwise.
    The structure definition includes the alignment required
    to produce field offsets specified in ``dtype``.

    Modules are cached and the function returns a single module instance for equal ``dtype``'s.
    Therefore inside a kernel it will be rendered with the same prefix everywhere it is used.
    This results in a behavior characteristic for a structural type system,
    same as for the basic dtype-ctype conversion.
    """
    if dtype.names is None:
        return ctype(dtype)
    else:
        if dtype not in _DTYPE_TO_CTYPE_MODULE:
            module = _get_struct_module(dtype)
            _DTYPE_TO_CTYPE_MODULE[dtype] = module
        else:
            module = _DTYPE_TO_CTYPE_MODULE[dtype]

        return module



"""
Test standard transformations
"""
import pytest

from reikna.cluda import Snippet
from reikna.pureparallel import PureParallel
import reikna.transformations as tr
from reikna.core import Parameter, Annotation, Type

from helpers import *


def pytest_generate_tests(metafunc):
    int_dtypes = [numpy.dtype('int32'), numpy.dtype('int64')]
    float_dtypes = [numpy.dtype('float32')]
    complex_dtypes = [numpy.dtype('complex64')]

    if 'any_dtype' in metafunc.funcargnames:
        dtypes = int_dtypes + float_dtypes + complex_dtypes
        metafunc.parametrize('any_dtype', dtypes, ids=[str(x) for x in dtypes])


def get_test_computation(arr_t):
    return PureParallel(
        [Parameter('output', Annotation(arr_t, 'o')),
        Parameter('input', Annotation(arr_t, 'i'))],
        "${output.store_idx}(${idxs[0]}, ${input.load_idx}(${idxs[0]}));")


def test_copy(some_thr, any_dtype):

    input_ = get_test_array((1000,), any_dtype)
    input_dev = some_thr.to_device(input_)
    output_dev = some_thr.empty_like(input_dev)

    test = get_test_computation(input_dev)
    copy = tr.copy(input_dev)

    test.parameter.input.connect(copy, copy.output, input_prime=copy.input)
    test.parameter.output.connect(copy, copy.input, output_prime=copy.output)
    testc = test.compile(some_thr)

    testc(output_dev, input_dev)
    assert diff_is_negligible(output_dev.get(), input_)


def test_scale_param(some_thr, any_dtype):

    input = get_test_array((1000,), any_dtype)
    p1 = get_test_array((1,), any_dtype)[0]
    p2 = get_test_array((1,), any_dtype)[0]
    input_dev = some_thr.to_device(input)
    output_dev = some_thr.empty_like(input_dev)

    test = get_test_computation(input_dev)
    scale = tr.scale_param(input_dev, any_dtype)

    test.parameter.input.connect(scale, scale.output, input_prime=scale.input, p1=scale.coeff)
    test.parameter.output.connect(scale, scale.input, output_prime=scale.output, p2=scale.coeff)
    testc = test.compile(some_thr)

    testc(output_dev, p1, input_dev, p2)
    assert diff_is_negligible(output_dev.get(), input * p1 * p2)


def test_scale_const(some_thr, any_dtype):

    input = get_test_array((1000,), any_dtype)
    p1 = get_test_array((1,), any_dtype)[0]
    p2 = get_test_array((1,), any_dtype)[0]
    input_dev = some_thr.to_device(input)
    output_dev = some_thr.empty_like(input_dev)

    test = get_test_computation(input_dev)
    scale1 = tr.scale_const(input_dev, p1)
    scale2 = tr.scale_const(input_dev, p2)

    test.parameter.input.connect(scale1, scale1.output, input_prime=scale1.input)
    test.parameter.output.connect(scale2, scale2.input, output_prime=scale2.output)
    testc = test.compile(some_thr)

    testc(output_dev, input_dev)
    assert diff_is_negligible(output_dev.get(), input * p1 * p2)


def test_split_combine_complex(some_thr):

    i1 = get_test_array((1000,), numpy.float32)
    i2 = get_test_array((1000,), numpy.float32)
    i1_dev = some_thr.to_device(i1)
    i2_dev = some_thr.to_device(i2)
    o1_dev = some_thr.empty_like(i1)
    o2_dev = some_thr.empty_like(i2)

    base_t = Type(numpy.complex64, shape=1000)
    test = get_test_computation(base_t)
    combine = tr.combine_complex(base_t)
    split = tr.split_complex(base_t)

    test.parameter.input.connect(combine, combine.output, i_real=combine.real, i_imag=combine.imag)
    test.parameter.output.connect(split, split.input, o_real=split.real, o_imag=split.imag)
    testc = test.compile(some_thr)

    testc(o1_dev, o2_dev, i1_dev, i2_dev)
    assert diff_is_negligible(o1_dev.get(), i1)
    assert diff_is_negligible(o2_dev.get(), i2)



import itertools

import pytest

import reikna.cluda as cluda
import reikna.cluda.vsize as vsize
from reikna.helpers import min_blocks, product
import reikna.cluda.dtypes as dtypes
from helpers import *

from pytest_threadgen import parametrize_thread_tuple, create_thread_in_tuple


vals_find_local_size = [
    ((40, 50), 1, (1, 1)),
    ((7, 11, 13), 1001, (7, 11, 13)),
    ((100,), 6, (6,)),
    ((3, 15), 12, (3, 4)),
    ((10, 10, 10, 10, 10), 32, (2, 2, 2, 2, 2)),
    ((16, 16, 128, 1024, 2), 30 * 32, (16, 4, 3, 5, 1))
    ]
@pytest.mark.parametrize(
    ('global_size', 'flat_local_size', 'expected_local_size'),
    vals_find_local_size,
    ids=[str(x[:2]) for x in vals_find_local_size])
def test_find_local_size(global_size, flat_local_size, expected_local_size):
    """
    Checking that ``find_local_size`` finds the sizes we expect from it.
    """
    local_size = vsize.find_local_size(global_size, flat_local_size)
    assert product(local_size) == flat_local_size
    assert local_size == expected_local_size


vals_group_dimensions = [
    ((1, 1, 1), (2, 2)),
    ((1, 2, 3), (2, 3)),
    ((16, 16), (7, 45)),
    ((16, 16), (7, 45, 50)),
    ((3, 4, 5), (72,)),
    ((3, 4, 5), (4, 5, 6))
    ]
@pytest.mark.parametrize(
    ('virtual_shape', 'available_shape'),
    vals_group_dimensions,
    ids=[str(x) for x in vals_group_dimensions])
def test_group_dimensions(virtual_shape, available_shape):
    """
    Tests that ``group_dimensions()`` obeys its contracts.
    """
    v_groups, a_groups = vsize.group_dimensions(virtual_shape, available_shape)
    v_dims = []
    a_dims = []
    for v_group, a_group in zip(v_groups, a_groups):
        v_shape = virtual_shape[v_group[0]:v_group[-1]+1]
        a_shape = available_shape[a_group[0]:a_group[-1]+1]
        assert(product(v_shape) <= product(a_shape))

        v_dims += v_group
        a_dims += a_group

    assert v_dims == list(range(len(virtual_shape)))
    assert a_dims == list(range(len(available_shape[:len(a_dims)])))


vals_find_bounding_shape = [
    (256, (15, 18)),
    (256, (10, 3, 10)),
    (299, (10, 3, 10))
    ]
@pytest.mark.parametrize(
    ('virtual_size', 'available_shape'),
    vals_find_bounding_shape,
    ids=[str(x) for x in vals_find_bounding_shape])
def test_find_bounding_shape(virtual_size, available_shape):
    """
    Tests that ``find_bounding_shape()`` obeys its contracts.
    """
    shape = vsize.find_bounding_shape(virtual_size, available_shape)
    assert all(isinstance(d, int) for d in shape)
    assert product(shape) >= virtual_size
    assert all(d <= ad for d, ad in zip(shape, available_shape))


def pytest_generate_tests(metafunc):
    if 'testvs' in metafunc.funcargnames:
        global_sizes = [
            (35,), (31*31*4,),
            (15, 13), (13, 35),
            (17, 15, 13), (5, 33, 75), (10, 10, 10, 5)]
        local_sizes = [None, (4,), (2, 4), (6, 4, 2)]
        mngs = [(26, 31), (34, 56, 25)]
        mwiss = [(4, 4), (5, 3), (9, 5, 3)]

        vals = []

        for gs, ls, mng, mwis in itertools.product(global_sizes, local_sizes, mngs, mwiss):
            testvs = TestVirtualSizes.try_create(gs, ls, mng, mwis)
            if testvs is None:
                continue
            vals.append(testvs)

        metafunc.parametrize('testvs', vals, ids=[str(x) for x in vals])

    if 'incorrect_testvs' in metafunc.funcargnames:
        vals = [
            # Bounding global size (32, 32) is too big for the grid limit and given block size
            TestVirtualSizes((32, 32), (4, 4), (7, 8), (4, 4)),
            # Local size is too big
            TestVirtualSizes((32, 32), (5, 4), (16, 16), (4, 4)),
            ]
        metafunc.parametrize('incorrect_testvs', vals, ids=[str(x) for x in vals])


class ReferenceIds:

    def __init__(self, global_size, local_size):
        self.global_size = global_size
        if local_size is not None:
            self.local_size = local_size
            self.grid_size = tuple(min_blocks(gs, ls) for gs, ls in zip(global_size, local_size))

    def _tile_pattern(self, pattern, axis, full_shape):

        pattern_shape = [x if i == axis else 1 for i, x in enumerate(full_shape)]
        pattern = pattern.reshape(*pattern_shape)

        tile_shape = [x if i != axis else 1 for i, x in enumerate(full_shape)]
        pattern = numpy.tile(pattern, tile_shape)

        return pattern.astype(numpy.int32)

    def predict_local_ids(self, dim):
        global_len = self.global_size[dim]
        local_len = self.local_size[dim]
        repetitions = min_blocks(global_len, local_len)

        pattern = numpy.tile(numpy.arange(local_len), repetitions)[:global_len]
        return self._tile_pattern(pattern, dim, self.global_size)

    def predict_group_ids(self, dim):
        global_len = self.global_size[dim]
        local_len = self.local_size[dim]
        repetitions = min_blocks(global_len, local_len)

        pattern = numpy.repeat(numpy.arange(repetitions), local_len)[:global_len]
        return self._tile_pattern(pattern, dim, self.global_size)

    def predict_global_ids(self, dim):
        global_len = self.global_size[dim]

        pattern = numpy.arange(global_len)
        return self._tile_pattern(pattern, dim, self.global_size)


class TestVirtualSizes:

    @classmethod
    def try_create(cls, global_size, local_size, max_num_groups, max_work_item_sizes):
        """
        This method is used to filter working combinations of parameters
        from the cartesian product of all possible ones.
        Returns ``None`` if the parameters are not compatible.
        """
        if len(max_num_groups) != len(max_work_item_sizes):
            return None

        if local_size is not None:
            if len(local_size) > len(global_size):
                return None
            else:
                # we need local size and global size of the same length
                local_size = local_size + (1,) * (len(global_size) - len(local_size))

            if product(local_size) > product(max_work_item_sizes):
                return None

            bounding_global_size = [
                ls * min_blocks(gs, ls) for gs, ls
                in zip(global_size, local_size)]

            if product(bounding_global_size) > product(max_num_groups):
                return None

        else:
            if product(global_size) > product(max_num_groups):
                return None

        return cls(global_size, local_size, max_num_groups, max_work_item_sizes)

    def __init__(self, global_size, local_size, max_num_groups, max_work_item_sizes):
        self.global_size = global_size
        self.local_size = local_size
        if local_size is not None:
            self.grid_size = tuple(min_blocks(gs, ls) for gs, ls in zip(global_size, local_size))

        self.max_num_groups = max_num_groups
        self.max_work_item_sizes = max_work_item_sizes

    def is_supported_by(self, thr):
        return (
            len(self.max_num_groups) <= len(thr.device_params.max_num_groups) and
            all(ng <= mng for ng, mng
                in zip(self.max_num_groups, thr.device_params.max_num_groups)) and
            len(self.max_work_item_sizes) <= len(thr.device_params.max_work_item_sizes) and
            all(wi <= mwi for wi, mwi
                in zip(self.max_work_item_sizes, thr.device_params.max_work_item_sizes)))

    def __str__(self):
        return "{gs}-{ls}-limited-by-{mng}-{mwis}".format(
            gs=self.global_size, ls=self.local_size,
            mng=self.max_num_groups, mwis=self.max_work_item_sizes)


class override_device_params:
    """
    Some of the tests here need to make thread/workgroup number limits
    in DeviceParameters of the thread lower, so that they are easier to test.

    This context manager hacks into the Thread and replaces the ``device_params`` attribute.
    Since threads are reused, the old device_params must be restored on exit.
    """

    def __init__(self, thr, **kwds):
        self._thr = thr
        self._kwds = kwds

    def __enter__(self):
        self._old_device_params = self._thr.device_params
        device_params = self._thr.api.DeviceParameters(self._thr._device)
        for kwd, val in self._kwds.items():
            setattr(device_params, kwd, val)
        self._thr.device_params = device_params
        return self._thr

    def __exit__(self, *args):
        self._thr.device_params = self._old_device_params


def test_ids(thr, testvs):
    """
    Test that virtual IDs are correct for each thread.
    """
    if not testvs.is_supported_by(thr):
        pytest.skip()

    ref = ReferenceIds(testvs.global_size, testvs.local_size)

    with override_device_params(
            thr, max_num_groups=testvs.max_num_groups,
            max_work_item_sizes=testvs.max_work_item_sizes, warp_size=2) as limited_thr:

        get_ids = limited_thr.compile_static("""
        KERNEL void get_ids(
            GLOBAL_MEM int *local_ids,
            GLOBAL_MEM int *group_ids,
            GLOBAL_MEM int *global_ids,
            int vdim)
        {
            VIRTUAL_SKIP_THREADS;
            const VSIZE_T i = virtual_global_flat_id();
            local_ids[i] = virtual_local_id(vdim);
            group_ids[i] = virtual_group_id(vdim);
            global_ids[i] = virtual_global_id(vdim);
        }
        """, 'get_ids', testvs.global_size, local_size=testvs.local_size)

    local_ids = thr.array(ref.global_size, numpy.int32)
    group_ids = thr.array(ref.global_size, numpy.int32)
    global_ids = thr.array(ref.global_size, numpy.int32)

    for vdim in range(len(testvs.global_size)):

        get_ids(local_ids, group_ids, global_ids, numpy.int32(vdim))

        assert diff_is_negligible(global_ids.get(), ref.predict_global_ids(vdim))
        if testvs.local_size is not None:
            assert diff_is_negligible(local_ids.get(), ref.predict_local_ids(vdim))
            assert diff_is_negligible(group_ids.get(), ref.predict_group_ids(vdim))


def test_sizes(thr, testvs):
    """
    Test that virtual sizes are correct.
    """
    if not testvs.is_supported_by(thr):
        pytest.skip()

    ref = ReferenceIds(testvs.global_size, testvs.local_size)
    vdims = len(testvs.global_size)

    with override_device_params(
        thr, max_num_groups=testvs.max_num_groups,
        max_work_item_sizes=testvs.max_work_item_sizes, warp_size=2) as limited_thr:

        get_sizes = limited_thr.compile_static("""
        KERNEL void get_sizes(GLOBAL_MEM int *sizes)
        {
            if (virtual_global_id(0) > 0) return;

            for (int i = 0; i < ${vdims}; i++)
            {
                sizes[i] = virtual_local_size(i);
                sizes[i + ${vdims}] = virtual_num_groups(i);
                sizes[i + ${vdims * 2}] = virtual_global_size(i);
            }
            sizes[${vdims * 3}] = virtual_global_flat_size();
        }
        """,
            'get_sizes',
            testvs.global_size, local_size=testvs.local_size,
            render_kwds=dict(vdims=vdims))

    sizes = thr.array(vdims * 3 + 1, numpy.int32)
    get_sizes(sizes)

    sizes = sizes.get()
    local_sizes = sizes[0:vdims]
    grid_sizes = sizes[vdims:vdims*2]
    global_sizes = sizes[vdims*2:vdims*3]
    flat_size = sizes[vdims*3]

    global_sizes_ref = numpy.array(testvs.global_size).astype(numpy.int32)
    assert diff_is_negligible(global_sizes, global_sizes_ref)
    assert flat_size == product(testvs.global_size)

    if testvs.local_size is not None:
        grid_sizes_ref = numpy.array(testvs.grid_size).astype(numpy.int32)
        assert diff_is_negligible(grid_sizes, grid_sizes_ref)
        local_sizes_ref = numpy.array(testvs.local_size).astype(numpy.int32)
        assert diff_is_negligible(local_sizes, local_sizes_ref)


def test_incorrect_sizes(thr, incorrect_testvs):
    """
    Test that for sizes which exceed the thread capability, the exception is raised.
    """
    with override_device_params(
        thr, max_num_groups=incorrect_testvs.max_num_groups,
        max_work_item_sizes=incorrect_testvs.max_work_item_sizes, warp_size=2) as limited_thr:

        with pytest.raises(ValueError):
            kernel = thr.compile_static("""
            KERNEL void test(GLOBAL_MEM int *temp)
            {
                temp[0] = 1;
            }
            """, 'test', incorrect_testvs.global_size, local_size=incorrect_testvs.local_size)



"""
This module contains a number of pre-created transformations.
"""

import reikna.cluda.dtypes as dtypes
import reikna.cluda.functions as functions
from reikna.core import Transformation, Parameter, Annotation, Type


def copy(arr_t, out_arr_t=None):
    """
    Returns an identity transformation (1 output, 1 input): ``output = input``.
    Output array type ``out_arr_t`` may have different strides,
    but must have the same shape and data type.
    """
    if out_arr_t is None:
        out_arr_t = arr_t
    else:
        if out_arr_t.shape != arr_t.shape or out_arr_t.dtype != arr_t.dtype:
            raise ValueError("Input and output arrays must have the same shape and data type")

    return Transformation(
        [Parameter('output', Annotation(out_arr_t, 'o')),
        Parameter('input', Annotation(arr_t, 'i'))],
        "${output.store_same}(${input.load_same});")


def scale_param(arr_t, coeff_dtype):
    """
    Returns a scaling transformation with dynamic parameter (1 output, 1 input, 1 scalar):
    ``output = input * coeff``.
    """
    return Transformation(
        [Parameter('output', Annotation(arr_t, 'o')),
        Parameter('input', Annotation(arr_t, 'i')),
        Parameter('coeff', Annotation(coeff_dtype))],
        "${output.store_same}(${mul}(${input.load_same}, ${coeff}));",
        render_kwds=dict(mul=functions.mul(arr_t.dtype, coeff_dtype, out_dtype=arr_t.dtype)))


def scale_const(arr_t, coeff):
    """
    Returns a scaling transformation with fixed parameter (1 output, 1 input):
    ``output = input * <coeff>``.
    """
    coeff_dtype = dtypes.detect_type(coeff)
    return Transformation(
        [Parameter('output', Annotation(arr_t, 'o')),
        Parameter('input', Annotation(arr_t, 'i'))],
        "${output.store_same}(${mul}(${input.load_same}, ${coeff}));",
        render_kwds=dict(
            mul=functions.mul(arr_t.dtype, coeff_dtype, out_dtype=arr_t.dtype),
            coeff=dtypes.c_constant(coeff, dtype=coeff_dtype)))


def split_complex(input_arr_t):
    """
    Returns a transformation which splits complex input into two real outputs
    (2 outputs, 1 input): ``real = Re(input), imag = Im(input)``.
    """
    output_t = Type(dtypes.real_for(input_arr_t.dtype), shape=input_arr_t.shape)
    return Transformation(
        [Parameter('real', Annotation(output_t, 'o')),
        Parameter('imag', Annotation(output_t, 'o')),
        Parameter('input', Annotation(input_arr_t, 'i'))],
        """
            ${real.store_same}(${input.load_same}.x);
            ${imag.store_same}(${input.load_same}.y);
        """)


def combine_complex(output_arr_t):
    """
    Returns a transformation which joins two real inputs into complex output
    (1 output, 2 inputs): ``output = real + 1j * imag``.
    """
    input_t = Type(dtypes.real_for(output_arr_t.dtype), shape=output_arr_t.shape)
    return Transformation(
        [Parameter('output', Annotation(output_arr_t, 'o')),
        Parameter('real', Annotation(input_t, 'i')),
        Parameter('imag', Annotation(input_t, 'i'))],
        """
        ${output.store_same}(
            COMPLEX_CTR(${output.ctype})(
                ${real.load_same},
                ${imag.load_same}));
        """)



import weakref
from collections import namedtuple

from reikna.helpers import Graph
from reikna.core.signature import Parameter, Annotation, Type, Signature
from reikna.core.transformation import TransformationTree, TransformationParameter


class ComputationParameter(Type):
    """
    Bases: :py:class:`~reikna.core.Type`

    Represents a typed computation parameter.
    Can be used as a substitute of an array for functions
    which are only interested in array metadata.
    """

    def __init__(self, computation, name, type_):
        """__init__()""" # hide the signature from Sphinx

        Type.__init__(self, type_.dtype, shape=type_.shape, strides=type_.strides)
        self._computation = weakref.ref(computation)
        self._name = name

    def belongs_to(self, comp):
        return self._computation() is comp

    def connect(self, _trf, _tr_connector, **tr_from_comp):
        """
        Shortcut for :py:meth:`~reikna.core.Computation.connect`
        with this parameter as a first argument.
        """
        return self._computation().connect(self._name, _trf, _tr_connector, **tr_from_comp)

    def __str__(self):
        return self._name


class Translator:
    """
    A callable that translates strings from the known list, and prefixes unknown ones.
    Used to introduce parameter names from a nested computation to the parent namespace.
    """

    def __init__(self, known_old, known_new, prefix):
        self._mapping = {old:new for old, new in zip(known_old, known_new)}
        self._prefix = prefix

    def __call__(self, name):
        if name in self._mapping:
            return self._mapping[name]
        else:
            return (self._prefix + '_' if self._prefix != '' else '') + name

    def get_nested(self, known_old, known_new, prefix):
        """Returns a new ``Translator`` with an extended prefix."""
        return Translator(known_old, known_new, prefix + self._prefix)

    @classmethod
    def identity(cls):
        return cls([], [], "")


def check_external_parameter_name(name):
    """
    Checks that a user-supplied parameter name meets the special criteria.
    Basically, we do not want such names start with underscores
    to prevent them from conflicting with internal names.
    """
    # Raising errors here so we can provide better explanation for the user
    if name.startswith('_'):
        raise ValueError("External parameter name cannot start with the underscore.")


def make_parameter_container(parent, parameters):
    """
    Creates a convenience object with ``ComputationParameter`` attributes.
    """
    params_container = namedtuple(
        'ComputationParameters', [param.name for param in parameters])
    param_objs = [
        ComputationParameter(parent, param.name, param.annotation.type)
        for param in parameters]
    return params_container(*param_objs)


class Computation:
    """
    A base class for computations, intended to be subclassed.

    :param root_parameters: a list of :py:class:`~reikna.core.Parameter` objects.

    .. py:attribute:: signature

        A :py:class:`~reikna.core.Signature` object representing current computation signature
        (taking into account connected transformations).

    .. py:attribute:: parameter

        A named tuple of :py:class:`~reikna.core.computation.ComputationParameter` objects
        corresponding to parameters from the current :py:attr:`signature`.
    """

    def __init__(self, root_parameters):
        for param in root_parameters:
            check_external_parameter_name(param.name)
        self._tr_tree = TransformationTree(root_parameters)
        self._update_attributes()

    def _update_attributes(self):
        """
        Updates ``signature`` and ``parameter`` attributes.
        Called by the methods that change the signature.
        """
        leaf_params = self._tr_tree.get_leaf_parameters()
        self.signature = Signature(leaf_params)
        self.parameter = make_parameter_container(self, leaf_params)

    # The names are underscored to avoid name conflicts with ``tr_from_comp`` keys
    # (where the user can introduce new parameter names)
    def connect(self, _comp_connector, _trf, _tr_connector, **tr_from_comp):
        """
        Connect a transformation to the computation.

        :param _comp_connector: connection target ---
            a :py:class:`~reikna.core.computation.ComputationParameter` object
            beloning to this computation object, or a string with its name.
        :param _trf: a :py:class:`~reikna.core.Transformation` object.
        :param _tr_connector: connector on the side of the transformation ---
            a :py:class:`~reikna.core.transformation.TransformationParameter` object
            beloning to ``tr``, or a string with its name.
        :param tr_from_comp: a dictionary with the names of new or old
            computation parameters as keys, and
            :py:class:`~reikna.core.transformation.TransformationParameter` objects
            (or their names) as values.
        :returns: this computation object (modified).

        .. note::

            The resulting parameter order is determined by traversing
            the graph of connections depth-first (starting from the initial computation parameters),
            with the additional condition: the nodes do not change their order
            in the same branching level (i.e. in the list of computation or
            transformation parameters, both of which are ordered).

            For example, consider a computation with parameters ``(a, b, c, d)``.
            If you connect a transformation ``(a', c) -> a``, the resulting computation
            will have the signature ``(a', b, c, d)`` (as opposed to ``(a', c, b, d)``
            it would have for the pure depth-first traversal).
        """

        # Extract connector name
        if isinstance(_comp_connector, ComputationParameter):
            if not _comp_connector.belongs_to(self):
                raise ValueError("The connection target must belong to this computation.")
        param_name = str(_comp_connector)

        # Extract transformation parameters names

        if param_name in tr_from_comp:
            raise ValueError(
                "Parameter '" + param_name + "' cannot be supplied " +
                "both as the main connector and one of the child connections")

        tr_from_comp[param_name] = _tr_connector
        comp_from_tr = {}
        for comp_connection_name, tr_connection in tr_from_comp.items():
            check_external_parameter_name(comp_connection_name)
            if isinstance(tr_connection, TransformationParameter):
                if not tr_connection.belongs_to(_trf):
                    raise ValueError(
                        "The transformation parameter must belong to the provided transformation")
            tr_connection_name = str(tr_connection)
            comp_from_tr[tr_connection_name] = comp_connection_name

        self._tr_tree.connect(param_name, _trf, comp_from_tr)
        self._update_attributes()
        return self

    def _translate_tree(self, translator):
        return self._tr_tree.translate(translator)

    def _get_plan(self, tr_tree, translator, thread, fast_math):
        plan_factory = lambda: ComputationPlan(tr_tree, translator, thread, fast_math)
        args = [
            KernelArgument(param.name, param.annotation.type)
            for param in tr_tree.get_root_parameters()]
        return self._build_plan(plan_factory, thread.device_params, *args)

    def compile(self, thread, fast_math=False):
        """
        Compiles the computation with the given :py:class:`~reikna.cluda.api.Thread` object
        and returns a :py:class:`~reikna.core.computation.ComputationCallable` object.
        If ``fast_math`` is enabled, the compilation of all kernels is performed using
        the compiler options for fast and imprecise mathematical functions.
        """
        translator = Translator.identity()
        return self._get_plan(self._tr_tree, translator, thread, fast_math).finalize()

    def _build_plan(self, plan_factory, device_params, *args):
        """
        Derived classes override this method.
        It is called by :py:meth:`compile` and
        supposed to return a :py:class:`~reikna.core.computation.ComputationPlan` object.

        :param plan_factory: a callable returning a new
            :py:class:`~reikna.core.computation.ComputationPlan` object.
        :param device_params: a :py:class:`~reikna.cluda.api.DeviceParameters` object corresponding
            to the thread the computation is being compiled for.
        :param args: :py:class:`~reikna.core.computation.KernelArgument` objects,
            corresponding to ``parameters`` specified during the creation
            of this computation object.
        """
        raise NotImplementedError


class IdGen:
    """
    Encapsulates a simple ID generator.
    """

    def __init__(self, prefix, counter=0):
        self._counter = counter
        self._prefix = prefix

    def __call__(self):
        self._counter += 1
        return self._prefix + str(self._counter)


class KernelArgument(Type):
    """
    Bases: :py:class:`~reikna.core.Type`

    Represents an argument suitable to pass to planned kernel or computation call.
    """

    def __init__(self, name, type_):
        """__init__()""" # hide the signature from Sphinx
        Type.__init__(self, type_.dtype, shape=type_.shape, strides=type_.strides)
        self.name = name

    def __repr__(self):
        return "KernelArgument(" + self.name + ")"


class ComputationPlan:
    """
    Computation plan recorder.
    """

    def __init__(self, tr_tree, translator, thread, fast_math):
        """__init__()""" # hide the signature from Sphinx

        self._thread = thread
        self._tr_tree = tr_tree
        self._translator = translator
        self._fast_math = fast_math

        self._nested_comp_idgen = IdGen('_nested')
        self._persistent_value_idgen = IdGen('_value')
        self._temp_array_idgen = IdGen('_temp')

        self._external_annotations = self._tr_tree.get_root_annotations()
        self._persistent_values = {}
        self._temp_arrays = set()
        self._internal_annotations = {}
        self._kernels = []

    def _scalar(self, val):
        """
        Adds a persistent scalar to the plan, and returns the corresponding
        :py:class:`KernelArgument`.
        """
        name = self._translator(self._persistent_value_idgen())
        ann = Annotation(val)
        self._internal_annotations[name] = ann
        self._persistent_values[name] = ann.type(val)
        return KernelArgument(name, ann.type)

    def persistent_array(self, arr):
        """
        Adds a persistent GPU array to the plan, and returns the corresponding
        :py:class:`KernelArgument`.
        """
        name = self._translator(self._persistent_value_idgen())
        ann = Annotation(arr, 'i')
        self._internal_annotations[name] = ann
        self._persistent_values[name] = self._thread.to_device(arr)
        return KernelArgument(name, ann.type)

    def temp_array(self, shape, dtype, strides=None):
        """
        Adds a temporary GPU array to the plan, and returns the corresponding
        :py:class:`KernelArgument`.
        Temporary arrays can share physical memory, but in such a way that
        their contents is guaranteed to persist between the first and the last use in a kernel
        during the execution of the plan.
        """
        name = self._translator(self._temp_array_idgen())
        ann = Annotation(Type(dtype, shape=shape, strides=strides), 'io')
        self._internal_annotations[name] = ann
        self._temp_arrays.add(name)
        return KernelArgument(name, ann.type)

    def temp_array_like(self, arr):
        """
        Same as :py:meth:`temp_array`, taking the array properties
        from array or array-like object ``arr``.
        """
        return self.temp_array(arr.shape, arr.dtype, strides=arr.strides)

    def _get_annotation(self, name):
        if name in self._external_annotations:
            return self._external_annotations[name]
        else:
            return self._internal_annotations[name]

    def _process_kernel_arguments(self, args):
        """
        Scan through kernel arguments passed by the user, check types,
        and wrap ad hoc values if necessary.

        Does not change the plan state.
        """
        processed_args = []
        adhoc_idgen = IdGen('_adhoc')
        adhoc_values = {}

        for arg in args:
            if not isinstance(arg, KernelArgument):
                if hasattr(arg, 'shape') and hasattr(arg, 'dtype'):
                    if len(arg.shape) > 0:
                        raise ValueError("Arrays are not allowed as ad hoc arguments")

                    # Not creating a new persistent scalar with _scalar(),
                    # because the kernel compilation may fail,
                    # in which case we would have to roll back the plan state.
                    # These arguments are local to this kernel anyway,
                    # so there's no need in registering them in the plan.
                    name = self._translator(adhoc_idgen())
                    adhoc_values[name] = arg
                    annotation = Annotation(Type(arg.dtype))
                    arg = KernelArgument(name, annotation.type)
                else:
                    raise TypeError("Unknown argument type: " + str(type(arg)))
            else:
                annotation = self._get_annotation(arg.name)

            processed_args.append(Parameter(arg.name, annotation))

        return processed_args, adhoc_values

    def _process_computation_arguments(self, signature, args, kwds):
        """
        Scan through nested computation arguments passed by the user, check types,
        and create new persistent values for ad hoc arguments.

        Changes the plan state.
        """
        bound_args = signature.bind_with_defaults(args, kwds, cast=False)

        args = []
        for arg, param in zip(bound_args.args, signature.parameters.values()):

            if not isinstance(arg, KernelArgument):
                if param.annotation.array:
                    raise ValueError("Ad hoc arguments are only allowed for scalar parameters")
                arg = self._scalar(param.annotation.type(arg))

            annotation = self._get_annotation(arg.name)

            if not annotation.can_be_argument_for(param.annotation):
                raise TypeError(
                    "Got {annotation} for '{name}', expected {param_annotation}".format(
                        annotation=annotation, name=param.name,
                        param_annotation=param.annotation))

            args.append(arg)

        return [arg.name for arg in args]

    def kernel_call(self, template_def, args, global_size,
            local_size=None, render_kwds=None):
        """
        Adds a kernel call to the plan.

        :param template_def: Mako template def for the kernel.
        :param args: a list consisting of
            :py:class:`~reikna.core.computation.KernelArgument` objects,
            or scalar values wrapped in ``numpy.ndarray``,
            that are going to be passed to the kernel during execution.
        :param global_size: global size to use for the call, in **row-major** order.
        :param local_size: local size to use for the call, in **row-major** order.
            If ``None``, the local size will be picked automatically.
        :param render_kwds: dictionary with additional values used to render the template.
        """

        processed_args, adhoc_values = self._process_kernel_arguments(args)
        subtree = self._tr_tree.get_subtree(processed_args)

        kernel_name = '_kernel_func'
        kernel_declaration, kernel_leaf_names = subtree.get_kernel_declaration(kernel_name)
        kernel_argobjects = subtree.get_kernel_argobjects()

        if render_kwds is None:
            render_kwds = {}
        else:
            render_kwds = dict(render_kwds)

        kernel = self._thread.compile_static(
            template_def, kernel_name, global_size, local_size=local_size,
            render_args=[kernel_declaration] + kernel_argobjects,
            render_kwds=render_kwds,
            fast_math=self._fast_math)

        self._kernels.append(PlannedKernelCall(kernel, kernel_leaf_names, adhoc_values))

    def computation_call(self, computation, *args, **kwds):
        """
        Adds a nested computation call.
        The ``computation`` value must be a :py:class:`~reikna.core.Computation` object.
        ``args`` and ``kwds`` are values to be passed to the computation.
        """

        signature = computation.signature
        argnames = self._process_computation_arguments(signature, args, kwds)

        # We want to preserve the Computation object for which we're calling _build_plan()
        # (because it may have attributes created by the constructor of the derived class),
        # but we need to translate its tree to integrate names of its nodes into
        # the parent namespace.
        translator = self._translator.get_nested(
            signature.parameters, argnames, self._nested_comp_idgen())
        new_tree = computation._translate_tree(translator)
        new_tree.reconnect(self._tr_tree)

        self._append_plan(computation._get_plan(
            new_tree, translator, self._thread, self._fast_math))

    def _append_plan(self, plan):
        self._kernels += plan._kernels
        self._persistent_values.update(plan._persistent_values)
        self._temp_arrays.update(plan._temp_arrays)
        self._internal_annotations.update(plan._internal_annotations)

    def finalize(self):

        # We need to add inferred dependencies between temporary buffers.
        # Basically, we assume that if some buffer X was used first in kernel M
        # and last in kernel N, all buffers in kernels from M+1 till N-1 depend on it
        # (in other words, data in X has to persist from call M till call N)

        # Record first and last kernel when a certain temporary array was used.
        usage = {}
        for i, kernel in enumerate(self._kernels):
            for argname in kernel.argnames:
                if argname not in self._temp_arrays:
                    continue
                if argname in usage:
                    usage[argname][1] = i
                else:
                    usage[argname] = [i, i]

        # Convert usage to dependencies.
        # Basically, if usage ranges of two arrays overlap, they are dependent.
        dependencies = Graph()
        for name, pair in usage.items():
            start, end = pair
            for i in range(start, end + 1):
                for other_name in self._kernels[i].argnames:
                    if other_name not in self._temp_arrays or other_name == name:
                        continue
                    dependencies.add_edge(name, other_name)

        # Create a dictionary of internal values to be used by kernels.

        internal_args = dict(self._persistent_values)

        # Allocate buffers specifying the dependencies
        all_buffers = []
        for name in self._temp_arrays:
            dependent_buffers = []
            for dep in dependencies[name]:
                if dep in internal_args:
                    dependent_buffers.append(internal_args[dep])

            type_ = self._internal_annotations[name].type
            new_buf = self._thread.temp_array(
                type_.shape, type_.dtype, strides=type_.strides, dependencies=dependent_buffers)
            internal_args[name] = new_buf
            all_buffers.append(new_buf)

        return ComputationCallable(
            self._tr_tree.get_leaf_parameters(),
            self._kernels,
            internal_args,
            all_buffers)


class PlannedKernelCall:

    def __init__(self, kernel, argnames, adhoc_values):
        self._kernel = kernel
        self.argnames = argnames
        self._adhoc_values = adhoc_values

    def finalize(self, known_args):
        args = [None] * len(self.argnames)
        external_arg_positions = []

        for i, name in enumerate(self.argnames):
            if name in known_args:
                args[i] = known_args[name]
            elif name in self._adhoc_values:
                args[i] = self._adhoc_values[name]
            else:
                external_arg_positions.append((name, i))

        return KernelCall(self._kernel, self.argnames, args, external_arg_positions)


class ComputationCallable:
    """
    A result of calling :py:meth:`~reikna.core.Computation.compile` on a computation.
    Represents a callable opaque GPGPU computation.

    .. py:attribute:: signature

        A :py:class:`~reikna.core.Signature` object.

    .. py:attribute:: parameter

        A named tuple of :py:class:`~reikna.core.Type` objects corresponding
        to the callable's parameters.
    """

    def __init__(self, parameters, kernel_calls, internal_args, temp_buffers):
        self.signature = Signature(parameters)
        self.parameter = make_parameter_container(self, parameters)
        self._kernel_calls = [kernel_call.finalize(internal_args) for kernel_call in kernel_calls]
        self._internal_args = internal_args
        self.__tempalloc__ = temp_buffers

    def __call__(self, *args, **kwds):
        """
        Execute the computation.
        """
        bound_args = self.signature.bind_with_defaults(args, kwds, cast=True)
        for kernel_call in self._kernel_calls:
            kernel_call(bound_args.arguments)


class KernelCall:

    def __init__(self, kernel, argnames, args, external_arg_positions):
        self._argnames = argnames # primarily for debugging purposes
        self._kernel = kernel
        self._args = args
        self._external_arg_positions = external_arg_positions

    def __call__(self, external_args):
        for name, pos in self._external_arg_positions:
            self._args[pos] = external_args[name]

        self._kernel(*self._args)

        # releasing references to arrays
        for name, pos in self._external_arg_positions:
            self._args[pos] = None



import numpy
import numpy.linalg as la
import reikna.cluda as cluda

api = cluda.ocl_api()
platform = api.get_platforms()[0]
device = platform.get_devices()[1]
thr = api.Thread(device)

src = """
KERNEL void test(GLOBAL_MEM float *dest)
{
    const int idx = get_global_id(0);
    dest[idx] = log((idx + 1e-6) / 100.);
}
"""



prg_false = thr.compile(src, fast_math=False)
prg_true = thr.compile(src, fast_math=True)

test_false = prg_false.test
test_true = prg_true.test

N = 100
n_dev = thr.array(N, numpy.float32)

test_false(n_dev, global_size=N)
ns_false = n_dev.get()

test_true(n_dev, global_size=N)
ns_true = n_dev.get()

print ns_false.ravel()
print ns_true.ravel()
print la.norm(ns_false - ns_true) / la.norm(ns_true)




"""
This module contains information about the library version.

.. py:attribute:: version

    A tuple with version numbers, major components first.

.. py:attribute:: full_version

    A string fully identifying the current build.

.. py:attribute:: git_revision

    A string with Git SHA identifying the revision used to create this build.

.. py:attribute:: release

    A boolean variable, equals ``True`` if current version is a release version.

"""
# THIS FILE IS GENERATED BY SETUP.PY


version = (0, 4, 0)
full_version = "0.4.0"
git_revision = "d3d3f5e65ea532380b560f3dccc1e2f1332710fb"
release = True



import sys
major, minor, _, _, _ = sys.version_info
if not ((major == 2 and minor >= 6) or (major == 3 and minor >= 2)):
    print("Python >=2.6 or >=3.2 is required to use this module.")
    sys.exit(1)

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

import os.path
import subprocess

def cd_to_project_root():
    project_root = os.path.split(os.path.realpath(__file__))[0]

    class CwdContext:
        def __enter__(self):
            self.old_path = os.getcwd()
            os.chdir(project_root)
            sys.path.insert(0, project_root)
        def __exit__(self, *args):
            del sys.path[0]
            os.chdir(self.old_path)

    return CwdContext()

def git_revision():
    """Get current git revision. Taken from numpy."""

    def _minimal_ext_cmd(cmd):
        # construct minimal environment
        env = {}
        for k in ['SYSTEMROOT', 'PATH']:
            v = os.environ.get(k)
            if v is not None:
                env[k] = v
        # LANGUAGE is used on win32
        env['LANGUAGE'] = 'C'
        env['LANG'] = 'C'
        env['LC_ALL'] = 'C'
        out = subprocess.Popen(cmd, stdout = subprocess.PIPE).communicate()[0]
        return out

    try:
        out = _minimal_ext_cmd(['git', 'rev-parse', 'HEAD'])
        revision = out.strip().decode('ascii')
    except OSError:
        revision = None

    return revision

def ensure_version_py():
    """
    Create contents for version.py. Idea taken from numpy.
    Returns version string to be passed to setup.py.
    """

    template = '''"""
This module contains information about the library version.

.. py:attribute:: version

    A tuple with version numbers, major components first.

.. py:attribute:: full_version

    A string fully identifying the current build.

.. py:attribute:: git_revision

    A string with Git SHA identifying the revision used to create this build.

.. py:attribute:: release

    A boolean variable, equals ``True`` if current version is a release version.

"""
# THIS FILE IS GENERATED BY SETUP.PY


version = {version}
full_version = "{full_version}"
git_revision = "{git_revision}"
release = {released}
'''

    VERSION = (0, 4, 0)
    RELEASED = True

    version_str = '%d.%d.%d' % VERSION
    full_version_str = version_str
    version_py_path = os.path.join('reikna', 'version.py')

    with cd_to_project_root():
        if os.path.exists('.git'):
            revision = git_revision()
        elif os.path.exists(version_py_path):
            # must be a source distribution, use existing version file
            globals_dict = {}
            with open(version_py_path) as f:
                exec(compile(f.read(), version_py_path, 'exec'), globals_dict)
            revision = globals_dict['git_revision']
        else:
            revision = None

    if revision is None:
        revision = "unknown"

    if not RELEASED:
        full_version_str += '+dev.' + revision[:7]

    contents = template.format(version=repr(VERSION),
        full_version=full_version_str, git_revision=revision, released=repr(RELEASED))

    with cd_to_project_root(), open(version_py_path, 'w') as f:
        f.write(contents)

    return version_str, full_version_str

if __name__ == '__main__':

    DOCUMENTATION = open('README.rst').read()
    _, VERSION_STR = ensure_version_py()

    # seuptool's ``install_requires`` and distutil's ``requires`` have slightly different format
    dependencies = [
        ('mako', '>= 0.8.0'),
        ('numpy', '>= 1.6.0'),
        ('funcsigs', '>= 0.3')]
    requires = list(map(lambda nr: nr[0] + '(' + nr[1] + ')', dependencies))
    install_requires = list(map(lambda nr: nr[0] + ' ' + nr[1], dependencies))

    setup(
        name='reikna',
        packages=['reikna', 'reikna/cluda', 'reikna/core', 'reikna/helpers', 'reikna/cbrng'],
        provides=['reikna'],
        requires=requires,
        install_requires=install_requires,
        extras_require=dict(
            pyopencl=["pyopencl>=2013.1"],
            pycuda=["pycuda>=2013.1"],
            test=["pytest>=2.3", "pytest-cov"]),
        package_data={'reikna': ['*.mako'],
                      'reikna/cluda': ['*.mako'],
                      'reikna/cbrng': ['*.mako']},
        version=VERSION_STR,
        author='Bogdan Opanchuk',
        author_email='bogdan@opanchuk.net',
        url='http://github.com/Manticore/reikna',
        description='GPGPU algorithms for PyCUDA and PyOpenCL',
        long_description=DOCUMENTATION,
        classifiers=[
            'Development Status :: 4 - Beta',
            'Intended Audience :: Science/Research',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: MIT License',
            'Programming Language :: Python',
            'Programming Language :: Python :: 2',
            'Programming Language :: Python :: 2.6',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.2',
            'Programming Language :: Python :: 3.3',
            'Topic :: Software Development',
            'Topic :: Scientific/Engineering',
            'Operating System :: OS Independent'
        ]
    )



import numpy
import numpy.linalg as la
import reikna.cluda as cluda

api = cluda.ocl_api()
platform = api.get_platforms()[0]
device = platform.get_devices()[1]
thr = api.Thread(device)

src = """
WITHIN_KERNEL VSIZE_T virtual_local_id(unsigned int dim)
{
    if (dim == 0)
    {
        SIZE_T flat_id =
            get_local_id(0) * 1 +
            0;

        return (flat_id / 1);
    }

    return 0;
}

WITHIN_KERNEL VSIZE_T virtual_local_size(unsigned int dim)
{
    if (dim == 0)
    {
        return 100;
    }

    return 1;
}

WITHIN_KERNEL VSIZE_T virtual_group_id(unsigned int dim)
{
    if (dim == 0)
    {
        SIZE_T flat_id =
            get_group_id(0) * 1 +
            0;

        return (flat_id / 1);
    }

    return 0;
}

WITHIN_KERNEL VSIZE_T virtual_num_groups(unsigned int dim)
{
    if (dim == 0)
    {
        return 1;
    }

    return 1;
}

WITHIN_KERNEL VSIZE_T virtual_global_id(unsigned int dim)
{
    return virtual_local_id(dim) + virtual_group_id(dim) * virtual_local_size(dim);
}

WITHIN_KERNEL VSIZE_T virtual_global_size(unsigned int dim)
{
    if(dim == 0)
    {
        return 100;
    }

    return 1;
}

WITHIN_KERNEL VSIZE_T virtual_global_flat_id()
{
    return
        virtual_global_id(0) * 1 +
        0;
}

WITHIN_KERNEL VSIZE_T virtual_global_flat_size()
{
    return
        virtual_global_size(0) *
        1;
}


WITHIN_KERNEL bool virtual_skip_local_threads()
{

    return false;
}

WITHIN_KERNEL bool virtual_skip_groups()
{

    return false;
}

WITHIN_KERNEL bool virtual_skip_global_threads()
{

    return false;
}

#define VIRTUAL_SKIP_THREADS if(virtual_skip_local_threads() || virtual_skip_groups() || virtual_skip_global_threads()) return

typedef struct {
    unsigned long  v[2];
}  _module0_;




            WITHIN_KERNEL _module0_ _module1_key_from_int(int idx)
            {
                _module0_ result;

                result.v[0] = 8618484947917433856
                ;
                result.v[1] = 0
                    + idx
                ;

                return result;
            }




typedef struct {
    unsigned long  v[4];
}  _module2_;




WITHIN_KERNEL _module2_ _module3_make_counter_from_int(int x)
{
    _module2_ result;
    result.v[0] = 0;
    result.v[1] = 0;
    result.v[2] = 0;
    result.v[3] = x;
    return result;
}


WITHIN_KERNEL INLINE unsigned long _module3_mulhilo(
    unsigned long *hip, unsigned long a, unsigned long b)
{
#ifdef CUDA
    *hip = __umul64hi(a, b);
#else
    *hip = mul_hi(a, b);
#endif
    return a * b;
}

WITHIN_KERNEL _module2_ _module3_bijection(
    const _module0_ key, const _module2_ counter)
{
    _module2_ X = counter;

    unsigned long key0 = key.v[0];
    unsigned long key1 = key.v[1];

    unsigned long hi0, lo0, hi1, lo1;

    // round 0

        lo0 = _module3_mulhilo(&hi0, 15197193596820024467, X.v[0]);
        lo1 = _module3_mulhilo(&hi1, 14581110107779764567, X.v[2]);
        X.v[0] = hi1 ^ X.v[1] ^ key0;
        X.v[1] = lo1;
        X.v[2] = hi0 ^ X.v[3] ^ key1;
        X.v[3] = lo0;

    // bump key
    key0 += 11400714819323198485;
    key1 += 13503953896175478587;

    // round 1

        lo0 = _module3_mulhilo(&hi0, 15197193596820024467, X.v[0]);
        lo1 = _module3_mulhilo(&hi1, 14581110107779764567, X.v[2]);
        X.v[0] = hi1 ^ X.v[1] ^ key0;
        X.v[1] = lo1;
        X.v[2] = hi0 ^ X.v[3] ^ key1;
        X.v[3] = lo0;

    // bump key
    key0 += 11400714819323198485;
    key1 += 13503953896175478587;

    // round 2

        lo0 = _module3_mulhilo(&hi0, 15197193596820024467, X.v[0]);
        lo1 = _module3_mulhilo(&hi1, 14581110107779764567, X.v[2]);
        X.v[0] = hi1 ^ X.v[1] ^ key0;
        X.v[1] = lo1;
        X.v[2] = hi0 ^ X.v[3] ^ key1;
        X.v[3] = lo0;

    // bump key
    key0 += 11400714819323198485;
    key1 += 13503953896175478587;

    // round 3

        lo0 = _module3_mulhilo(&hi0, 15197193596820024467, X.v[0]);
        lo1 = _module3_mulhilo(&hi1, 14581110107779764567, X.v[2]);
        X.v[0] = hi1 ^ X.v[1] ^ key0;
        X.v[1] = lo1;
        X.v[2] = hi0 ^ X.v[3] ^ key1;
        X.v[3] = lo0;

    // bump key
    key0 += 11400714819323198485;
    key1 += 13503953896175478587;

    // round 4

        lo0 = _module3_mulhilo(&hi0, 15197193596820024467, X.v[0]);
        lo1 = _module3_mulhilo(&hi1, 14581110107779764567, X.v[2]);
        X.v[0] = hi1 ^ X.v[1] ^ key0;
        X.v[1] = lo1;
        X.v[2] = hi0 ^ X.v[3] ^ key1;
        X.v[3] = lo0;

    // bump key
    key0 += 11400714819323198485;
    key1 += 13503953896175478587;

    // round 5

        lo0 = _module3_mulhilo(&hi0, 15197193596820024467, X.v[0]);
        lo1 = _module3_mulhilo(&hi1, 14581110107779764567, X.v[2]);
        X.v[0] = hi1 ^ X.v[1] ^ key0;
        X.v[1] = lo1;
        X.v[2] = hi0 ^ X.v[3] ^ key1;
        X.v[3] = lo0;

    // bump key
    key0 += 11400714819323198485;
    key1 += 13503953896175478587;

    // round 6

        lo0 = _module3_mulhilo(&hi0, 15197193596820024467, X.v[0]);
        lo1 = _module3_mulhilo(&hi1, 14581110107779764567, X.v[2]);
        X.v[0] = hi1 ^ X.v[1] ^ key0;
        X.v[1] = lo1;
        X.v[2] = hi0 ^ X.v[3] ^ key1;
        X.v[3] = lo0;

    // bump key
    key0 += 11400714819323198485;
    key1 += 13503953896175478587;

    // round 7

        lo0 = _module3_mulhilo(&hi0, 15197193596820024467, X.v[0]);
        lo1 = _module3_mulhilo(&hi1, 14581110107779764567, X.v[2]);
        X.v[0] = hi1 ^ X.v[1] ^ key0;
        X.v[1] = lo1;
        X.v[2] = hi0 ^ X.v[3] ^ key1;
        X.v[3] = lo0;

    // bump key
    key0 += 11400714819323198485;
    key1 += 13503953896175478587;

    // round 8

        lo0 = _module3_mulhilo(&hi0, 15197193596820024467, X.v[0]);
        lo1 = _module3_mulhilo(&hi1, 14581110107779764567, X.v[2]);
        X.v[0] = hi1 ^ X.v[1] ^ key0;
        X.v[1] = lo1;
        X.v[2] = hi0 ^ X.v[3] ^ key1;
        X.v[3] = lo0;

    // bump key
    key0 += 11400714819323198485;
    key1 += 13503953896175478587;

    // round 9

        lo0 = _module3_mulhilo(&hi0, 15197193596820024467, X.v[0]);
        lo1 = _module3_mulhilo(&hi1, 14581110107779764567, X.v[2]);
        X.v[0] = hi1 ^ X.v[1] ^ key0;
        X.v[1] = lo1;
        X.v[2] = hi0 ^ X.v[3] ^ key1;
        X.v[3] = lo0;



    return X;
}




typedef unsigned int _module3_uint32;
typedef unsigned long _module3_uint64;

typedef struct _module3_
{
    _module0_ key;
    _module2_ counter;
    union {
        _module2_ buffer;
        unsigned int buffer_uint32[8];
    };
    int buffer_uint32_cursor;
} _module3_STATE;


WITHIN_KERNEL void _module3_bump_counter(_module3_STATE *state)
{
    state->counter.v[3] += 1;
    if (state->counter.v[3] == 0)
    {
    state->counter.v[2] += 1;
    if (state->counter.v[2] == 0)
    {
    state->counter.v[1] += 1;
    if (state->counter.v[1] == 0)
    {
    state->counter.v[0] += 1;
    }
    }
    }
}

WITHIN_KERNEL void _module3_refill_buffer(_module3_STATE *state)
{
    state->buffer = _module3_bijection(state->key, state->counter);
}

WITHIN_KERNEL _module3_STATE _module3_make_state(_module0_ key, _module2_ counter)
{
    _module3_STATE state;
    state.key = key;
    state.counter = counter;
    state.buffer_uint32_cursor = 0;
    _module3_refill_buffer(&state);
    return state;
}

WITHIN_KERNEL unsigned int _module3_get_raw_uint32(_module3_STATE *state)
{
    if (state->buffer_uint32_cursor == 8)
    {
        _module3_bump_counter(state);
        state->buffer_uint32_cursor = 0;
        _module3_refill_buffer(state);
    }

    int cur = state->buffer_uint32_cursor;
    state->buffer_uint32_cursor += 1;
    return state->buffer_uint32[cur];
}

WITHIN_KERNEL unsigned long _module3_get_raw_uint64(_module3_STATE *state)
{
    if (state->buffer_uint32_cursor >= 8 - 1)
    {
        _module3_bump_counter(state);
        state->buffer_uint32_cursor = 0;
        _module3_refill_buffer(state);
    }

    int cur = state->buffer_uint32_cursor;
    state->buffer_uint32_cursor += 2;
    return state->buffer.v[cur / 2];
}





typedef float _module5_value;
#define _module5_RANDOMS_PER_CALL 1

typedef struct
{
    float v[1];
} _module5_RESULT;


WITHIN_KERNEL _module5_RESULT _module5_sample(_module3_STATE *state)
{
    _module5_RESULT result;
    float normalized = (float)_module3_get_raw_uint32(state) / 4294967296.0f;
    result.v[0] = normalized * (1.0f) + (0.0f);
    return result;
}




typedef float _module4_value;
#define _module4_RANDOMS_PER_CALL 2

typedef struct
{
    float v[2];
} _module4_RESULT;


WITHIN_KERNEL _module4_RESULT _module4_sample(_module3_STATE *state)
{
    _module4_RESULT result;
    _module5_RESULT r1 = _module5_sample(state);
    r1 = _module5_sample(state);
    _module5_RESULT r2 = _module5_sample(state);

    float u1 = r1.v[0];
    float u2 = r2.v[0];

    float ang = 6.2831854820251465f * u2;
    float c_ang = cos(ang);
    float s_ang = sin(ang);

    float coeff = sqrt(-2.0f * log(u1)) * (10.0f);

    result.v[0] = u1;
    result.v[1] = u1;
    return result;
}



        KERNEL void test(GLOBAL_MEM float *dest, int ctr_start)
        {
            VIRTUAL_SKIP_THREADS;
            const VSIZE_T idx = virtual_global_id(0);

            if (idx != 0)
                return;

            _module0_ key = _module1_key_from_int(idx);
            _module2_ ctr = _module3_make_counter_from_int(ctr_start);
            _module3_STATE st = _module3_make_state(key, ctr);

            _module4_RESULT res;

            for(int j = 0; j < 1; j++)
            {
                res = _module4_sample(&st);

                dest[j * 200 + 0 + idx] = res.v[0];
                dest[j * 200 + 100 + idx] = res.v[1];
            }
        }
"""



prg_false = thr.compile(src, fast_math=False)
prg_true = thr.compile(src, fast_math=True)

test_false = prg_false.test
test_true = prg_true.test

N = 100
n_dev = thr.to_device(numpy.zeros((2, N), numpy.float32))

test_false(n_dev, numpy.int32(0), global_size=N)
ns_false = n_dev.get()
print ns_false

test_true(n_dev, numpy.int32(0), global_size=N)
ns_true = n_dev.get()

print ns_true



print la.norm(ns_false - ns_true) / la.norm(ns_true)

